/*****************************************************************************/
/*                      WiNRADiO G35DDC EXCALIBUR ULTRA                      */
/*                                SDK DEMO                                   */
/*                      FFT and FFT overlapping buffer                       */
/*                        (C) Copyright RADIXON 2017                         */
/*****************************************************************************/

#include "u_fft.h"
#include <math.h>
#include <string.h>

//The original of the following FFT can be found on the following link:
//http://www.drdobbs.com/cpp/a-simple-and-efficient-fft-implementatio/199500857?pgno=1
//The function does in-place FFT
//data - complex signal - data[0]=real0, data[1]=img0, data[2]=real1, data[3]=img1, ...
//nn - FFT size
void FFT(float* data, unsigned long nn)
{
    unsigned long n, mmax, m, j, istep, i;
    double wtemp, wr, wpr, wpi, wi, theta;
    double tempr, tempi;
    float t;
 
    // reverse-binary reindexing
    n = nn<<1;
    j=1;
    for (i=1; i<n; i+=2) {
        if (j>i) {
            t=data[j-1];
            data[j-1]=data[i-1];
            data[i-1]=t;
            
            t=data[j];
            data[j]=data[i];
            data[i]=t;
        }
        m = nn;
        while (m>=2 && j>m) {
            j -= m;
            m >>= 1;
        }
        j += m;
    };
 
    // here begins the Danielson-Lanczos section
    mmax=2;
    while (n>mmax) {
        istep = mmax<<1;
        theta = -(2*M_PI/mmax);
        wtemp = sin(0.5*theta);
        wpr = -2.0*wtemp*wtemp;
        wpi = sin(theta);
        wr = 1.0;
        wi = 0.0;
        for (m=1; m < mmax; m += 2) {
            for (i=m; i <= n; i += istep) {
                j=i+mmax;
                tempr = wr*data[j-1] - wi*data[j];
                tempi = wr * data[j] + wi*data[j-1];
 
                data[j-1] = (float)(data[i-1] - tempr);
                data[j] = (float)(data[i] - tempi);
                data[i-1]=(float)(data[i-1]+tempr);
                data[i]=(float)(data[i]+tempi);
            }
            wtemp=wr;
            wr += wr*wpr - wi*wpi;
            wi += wi*wpr + wtemp*wpi;
        }
        mmax=istep;
    }
}

//Computes Blackman window function coeffs and normalizes them for FFT
//http://en.wikipedia.org/wiki/Window_function
void GetNormalizedWindowCoeffs(float *Buffer,int Count,double FixedCoeff)
{
 double a0,a1,a2;
 int n;
 double Sum;

    a0=(1-0.16)/2;
    a1=0.5;
    a2=0.16/2;
   
    Sum=0;

    for(n=0;n<Count;n++)
    {
        Buffer[n]=(float)(a0-a1*cos((2*M_PI*n)/(Count-1))+a2*cos((4*M_PI*n)/(Count-1)));
        Sum+=Buffer[n];
    }

    Sum=FixedCoeff/Sum;

    for(n=0;n<Count;n++)
    {
        Buffer[n]=(float)(Buffer[n]*Sum);
    }
}

TFFTBuffer::TFFTBuffer(uint32_t MaxSize)
{
    m_MaxSize=MaxSize;
    m_Buffer=new float[MaxSize*2];
    m_Size=1024;
    m_Used=0;
}

TFFTBuffer::~TFFTBuffer(void)
{
    delete[] m_Buffer;
}

bool TFFTBuffer::AddSamples_16(const int16_t *Samples,uint32_t Count)
{
 uint32_t i,j;
 uint32_t Remain,ToRemove;

    if(m_Size<=Count)
    {
        Samples+=(Count-m_Size)<<1;

        for(i=0;i<m_Size;i++)
        {
            m_Buffer[i*2]=(float)(Samples[i*2]*(1.0/32768.0));
            m_Buffer[i*2+1]=(float)(Samples[i*2+1]*(1.0/32768.0));
        }

        m_Used=m_Size;
        return true;
    }

    Remain=m_Size-m_Used;

    if(Remain<=Count)
    {
        ToRemove=Count-Remain;

        memmove(m_Buffer,m_Buffer+ToRemove*2,(m_Used-ToRemove)*2*sizeof(float));        

        for(i=m_Used-ToRemove,j=0;j<Count;i++,j++)
        {
            m_Buffer[i*2]=(float)(Samples[j*2]*(1.0/32768.0));
            m_Buffer[i*2+1]=(float)(Samples[j*2+1]*(1.0/32768.0));
        }

        m_Used=m_Size;
        return true;
    }

    for(i=m_Used,j=0;j<Count;i++,j++)
    {
        m_Buffer[i*2]=(float)(Samples[j*2]*(1.0/32768.0));
        m_Buffer[i*2+1]=(float)(Samples[j*2+1]*(1.0/32768.0));
    }

    m_Used+=Count;
    return false;
}

bool TFFTBuffer::AddSamples_32(const int32_t *Samples,uint32_t Count)
{
 uint32_t i,j;
 uint32_t Remain,ToRemove;

    if(m_Size<=Count)
    {
        Samples+=(Count-m_Size)<<1;

        for(i=0;i<m_Size;i++)
        {
            m_Buffer[i*2]=(float)(Samples[i*2]*(1.0/2147483648.0));
            m_Buffer[i*2+1]=(float)(Samples[i*2+1]*(1.0/2147483648.0));
        }

        m_Used=m_Size;
        return true;
    }

    Remain=m_Size-m_Used;

    if(Remain<=Count)
    {
        ToRemove=Count-Remain;

        memmove(m_Buffer,m_Buffer+ToRemove*2,(m_Used-ToRemove)*2*sizeof(float));

        for(i=m_Used-ToRemove,j=0;j<Count;i++,j++)
        {
            m_Buffer[i*2]=(float)(Samples[j*2]*(1.0/2147483648.0));
            m_Buffer[i*2+1]=(float)(Samples[j*2+1]*(1.0/2147483648.0));
        }

        m_Used=m_Size;
        return true;
    }

    for(i=m_Used,j=0;j<Count;i++,j++)
    {
        m_Buffer[i*2]=(float)(Samples[j*2]*(1.0/2147483648.0));
        m_Buffer[i*2+1]=(float)(Samples[j*2+1]*(1.0/2147483648.0));
    }

    m_Used+=Count;
    return false;
}

bool TFFTBuffer::AddSamples_F32(const float *Samples,uint32_t Count)
{
 uint32_t i,j;
 uint32_t Remain,ToRemove;

    if(m_Size<=Count)
    {
        Samples+=(Count-m_Size)<<1;

        for(i=0;i<m_Size;i++)
        {
            m_Buffer[i*2]=Samples[i*2];
            m_Buffer[i*2+1]=Samples[i*2+1];
        }

        m_Used=m_Size;
        return true;
    }

    Remain=m_Size-m_Used;

    if(Remain<=Count)
    {
        ToRemove=Count-Remain;

        memmove(m_Buffer,m_Buffer+ToRemove*2,(m_Used-ToRemove)*2*sizeof(float));

        for(i=m_Used-ToRemove,j=0;j<Count;i++,j++)
        {
            m_Buffer[i*2]=Samples[j*2];
            m_Buffer[i*2+1]=Samples[j*2+1];
        }

        m_Used=m_Size;
        return true;
    }

    for(i=m_Used,j=0;j<Count;i++,j++)
    {
        m_Buffer[i*2]=Samples[j*2];
        m_Buffer[i*2+1]=Samples[j*2+1];
    }

    m_Used+=Count;
    return false;
}

bool TFFTBuffer::AddSamples_Real_Mono_F32(const float *Samples,uint32_t Count)
{
 uint32_t i,j;
 uint32_t Remain,ToRemove;

    if(m_Size<=Count)
    {
        Samples+=(Count-m_Size);

        for(i=0;i<m_Size;i++)
        {
            m_Buffer[i]=Samples[i];
        }

        m_Used=m_Size;
        return true;
    }

    Remain=m_Size-m_Used;

    if(Remain<=Count)
    {
        ToRemove=Count-Remain;

        memmove(m_Buffer,m_Buffer+ToRemove,(m_Used-ToRemove)*sizeof(float));

        for(i=m_Used-ToRemove,j=0;j<Count;i++,j++)
        {
            m_Buffer[i]=Samples[j];
        }

        m_Used=m_Size;
        return true;
    }

    for(i=m_Used,j=0;j<Count;i++,j++)
    {
        m_Buffer[i]=Samples[j];
    }

    m_Used+=Count;
    return false;
}