/* ***********************************************************************/
/*                                                                       */
/*                      WiNRADiO G35DDC EXCELSIOR                        */
/*                 G35DDC device driver (PCIe)                           */
/*                     (C) Copyright 2014 RADIXON.                       */
/*                                                                       */
/* ***********************************************************************/

#include "g35ddc.h"

static void g35_io_buffer_unmap(struct g35_io_buffer *io_buffer)
{
 int i;

    if(io_buffer->pages)
    {
        if(io_buffer->changed)
        {
            for(i=0;i<io_buffer->page_count;i++)
            {
                SetPageDirty(io_buffer->pages[i]);
                put_page(io_buffer->pages[i]);
            }
        }
        else
        {
            for(i=0;i<io_buffer->page_count;i++)
            {
                put_page(io_buffer->pages[i]);
            }
        }

        io_buffer->changed=0;

        kfree(io_buffer->pages);
        io_buffer->pages=NULL;
        io_buffer->page_count=0;
    }
}


static int g35_io_buffer_prepare(uintptr_t user_buffer,uint32_t size,int write,struct g35_io_buffer **io_buffer_r)
{
 struct g35_io_buffer *io_buffer;
 int page_count,locked_page_count,i;

    if(unlikely(user_buffer==0 || (user_buffer & (~PAGE_MASK))!=0 || size==0))
    {
        return -EINVAL;
    }

    page_count=(int)(((user_buffer+size-1)&PAGE_MASK)>>PAGE_SHIFT)-(int)(user_buffer>>PAGE_SHIFT)+1;

    io_buffer=kmalloc(sizeof(struct g35_io_buffer),GFP_KERNEL);

    if(unlikely(io_buffer==NULL))
    {
        return -ENOMEM;
    }

    io_buffer->pages=kmalloc(page_count*sizeof(struct page*),GFP_KERNEL);

    if(unlikely(io_buffer->pages==NULL))
    {
        kfree(io_buffer);
        return -ENOMEM;
    }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
    locked_page_count=get_user_pages_fast((uintptr_t)user_buffer,page_count,write,io_buffer->pages);
#else
    down_read(&current->mm->mmap_sem);
    locked_page_count=get_user_pages(current,current->mm,(uintptr_t)user_buffer,page_count,write,1,io_buffer->pages,NULL);
    up_read(&current->mm->mmap_sem);
#endif

    if(unlikely(locked_page_count<0))
    {
        kfree(io_buffer->pages);
        kfree(io_buffer);
        return locked_page_count;
    }

    if(unlikely(locked_page_count<page_count))
    {
        for(i=0;i<locked_page_count;i++)
        {
            put_page(io_buffer->pages[i]);
        }

        kfree(io_buffer->pages);
        kfree(io_buffer);
        return -ENOMEM;
    }

    io_buffer->user_buffer=user_buffer;
    io_buffer->size=size;
    io_buffer->page_count=locked_page_count;
    io_buffer->changed=0;

#ifdef __G35_PCI__
    io_buffer->sg_list=NULL;
    io_buffer->sg_segment_count=0;
    io_buffer->sg_pci_dev=NULL;
#endif //__G35_PCI__

    *io_buffer_r=io_buffer;

    return 0;
}

#ifdef __G35_PCI__

inline static void g35_io_buffer_sg_sync_for_cpu(struct g35_io_buffer *io_buffer)
{
    dma_sync_sg_for_cpu(&io_buffer->sg_pci_dev->dev,io_buffer->sg_list,io_buffer->page_count,io_buffer->sg_direction);
}

inline static void g35_io_buffer_sg_sync_for_device(struct g35_io_buffer *io_buffer)
{
    dma_sync_sg_for_device(&io_buffer->sg_pci_dev->dev,io_buffer->sg_list,io_buffer->page_count,io_buffer->sg_direction);
}

static int g35_io_buffer_sg_map(struct g35_io_buffer *io_buffer,struct pci_dev *pci_dev,enum dma_data_direction direction)
{
 struct scatterlist *sg_list;
 int i;
 int count;
 uint32_t size;

    sg_list=kcalloc(io_buffer->page_count,sizeof(*sg_list),GFP_KERNEL);

    if(unlikely(sg_list==NULL))
    {
        return -ENOMEM;
    }

    size=io_buffer->size;


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
     sg_init_table(sg_list,io_buffer->page_count);
#endif

    for(i=0;i<io_buffer->page_count;i++)
    {
        if(likely(size>=PAGE_SIZE))
        {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
            sg_set_page(&sg_list[i],io_buffer->pages[i],PAGE_SIZE,0);
#else
            sg_list[i].page=io_buffer->pages[i];
            sg_list[i].offset=0;
            sg_list[i].length=PAGE_SIZE;
#endif

            size-=PAGE_SIZE;
        }
        else
        {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
            sg_set_page(&sg_list[i],io_buffer->pages[i],size,0);
#else
            sg_list[i].page=io_buffer->pages[i];
            sg_list[i].offset=0;
            sg_list[i].length=size;
#endif

            size=0;
        }
    }

    count=dma_map_sg(&pci_dev->dev,sg_list,io_buffer->page_count,direction);

    if(unlikely(count==0))
    {
        kfree(sg_list);
        return -ENOMEM;
    }

    io_buffer->sg_list=sg_list;
    io_buffer->sg_segment_count=count;
    io_buffer->sg_pci_dev=pci_dev;
    io_buffer->sg_direction=direction;

    return 0;
}

static void g35_io_buffer_sg_unmap(struct g35_io_buffer *io_buffer)
{
    if(io_buffer->sg_list)
    {
        g35_io_buffer_sg_sync_for_cpu(io_buffer);

        dma_unmap_sg(&io_buffer->sg_pci_dev->dev,io_buffer->sg_list,io_buffer->page_count,io_buffer->sg_direction);
        kfree(io_buffer->sg_list);

        io_buffer->sg_list=NULL;
    }
}

#endif //__G35_PCI__

static void g35_io_buffer_free(struct g35_io_buffer *io_buffer)
{
 int i;

#ifdef __G35_PCI__
    g35_io_buffer_sg_unmap(io_buffer);
#endif //__G35_PCI__

    if(io_buffer->pages!=NULL)
    {
        if(io_buffer->changed)
        {
            for(i=0;i<io_buffer->page_count;i++)
            {
                if(!PageReserved(io_buffer->pages[i]))
                {
                    SetPageDirty(io_buffer->pages[i]);
                }

                put_page(io_buffer->pages[i]);
            }
        }
        else
        {
            for(i=0;i<io_buffer->page_count;i++)
            {
                put_page(io_buffer->pages[i]);
            }
        }

        kfree(io_buffer->pages);
    }

    kfree(io_buffer);
}

static int g35_io_stream_get_complete_buffer(struct g35ddc *dev,struct g35_io_stream *io_stream,uintptr_t *user_buffer)
{
 struct g35_io_buffer *io_buffer;
 unsigned long flags;
 struct g35_complete_io_buffers *buffers;

    buffers=&io_stream->complete_buffers;

    for(;;)
    {
        down_read(&io_stream->rw_semaphore);

        if(unlikely(!io_stream->active))
        {
            up_read(&io_stream->rw_semaphore);
            return -EPERM;
        }

        spin_lock_irqsave(&buffers->spinlock,flags);

        if(unlikely(list_empty(&buffers->list)))
        {
            io_buffer=NULL;
        }
        else
        {
            io_buffer=list_entry(buffers->list.next,struct g35_io_buffer,list_entry);

            list_del(&io_buffer->list_entry);
        }

        spin_unlock_irqrestore(&buffers->spinlock,flags);

        up_read(&io_stream->rw_semaphore);

        if(likely(io_buffer!=NULL))
        {
            break;
        }

        if(unlikely(down_interruptible(&buffers->semaphore)<0))
        {
            return -EINTR;
        }
    }

    *user_buffer=io_buffer->user_buffer;

    g35_io_buffer_free(io_buffer);

    return 0;
}

#ifdef __G35_PCI__

static struct g35_io_buffer* g35_io_stream_get_pending_buffer_nolock(struct g35_io_stream *io_stream)
{
 struct g35_pending_io_buffers *buffers;
 struct g35_io_buffer *io_buffer;

    buffers=&io_stream->pending_buffers;

    if(unlikely(list_empty(&buffers->list)))
    {
        io_buffer=NULL;
    }
    else
    {
        io_buffer=list_entry(buffers->list.next,struct g35_io_buffer,list_entry);

        list_del(buffers->list.next);
    }

    return io_buffer;
}

#endif //__G35_PCI__

static void g35_io_stream_put_complete_buffer(struct g35_io_stream *io_stream,struct g35_io_buffer *io_buffer)
{
 unsigned long flags;
 struct g35_complete_io_buffers *buffers;

    buffers=&io_stream->complete_buffers;

    g35_io_buffer_unmap(io_buffer);

    spin_lock_irqsave(&buffers->spinlock,flags);

    list_add_tail(&io_buffer->list_entry,&buffers->list);

    up(&buffers->semaphore);

    spin_unlock_irqrestore(&buffers->spinlock,flags);
}


inline static void g35_io_stream_put_pending_buffer_nolock(struct g35_io_stream *io_stream,struct g35_io_buffer *io_buffer)
{
    list_add_tail(&io_buffer->list_entry,&io_stream->pending_buffers.list);
}

static void g35_io_stream_free_buffers(struct g35_io_stream *io_stream)
{
 struct g35_io_buffer *io_buffer;

    while(!list_empty(&io_stream->pending_buffers.list))
    {
        io_buffer=list_entry(io_stream->pending_buffers.list.next,struct g35_io_buffer,list_entry);

        list_del(&io_buffer->list_entry);

        g35_io_buffer_free(io_buffer);
    }

    INIT_LIST_HEAD(&io_stream->pending_buffers.list);

    while(!list_empty(&io_stream->complete_buffers.list))
    {
        io_buffer=list_entry(io_stream->complete_buffers.list.next,struct g35_io_buffer,list_entry);

        list_del(&io_buffer->list_entry);

        g35_io_buffer_free(io_buffer);
    }

    INIT_LIST_HEAD(&io_stream->complete_buffers.list);
}

static dev_t g35_dev0;

static struct class g35_class=
{
    .name=G35_DEVICE_NAME
};

static struct g35ddc *g35_registered_devices[G35_MAX_DEVICE_COUNT]={0};

static struct g35_mutex g35_reg_mutex;

static int g35_set_power(struct g35ddc *dev,uint8_t power,uint8_t lock);

static int g35_fo_open(struct inode *node, struct file *file);
static int g35_fo_release(struct inode *node, struct file *file);
static int g35_fo_mmap(struct file *file,struct vm_area_struct *vma);

#if defined(HAVE_UNLOCKED_IOCTL) || LINUX_VERSION_CODE>=KERNEL_VERSION(5,0,0)
static long g35_fo_ioctl(struct file *file, unsigned int cmd, unsigned long arg);
#else
static int g35_fo_ioctl(struct inode *node,struct file *file, unsigned int cmd, unsigned long arg);
#endif

#if defined(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL)
static long g35_fo_ioctl_compat(struct file *file, unsigned int cmd, unsigned long arg);
#endif

static struct file_operations g35_file_operations=
{
    .open=g35_fo_open,
    .release=g35_fo_release,
#if defined(HAVE_UNLOCKED_IOCTL) || LINUX_VERSION_CODE>=KERNEL_VERSION(5,0,0)
    .unlocked_ioctl=g35_fo_ioctl,
#else
    .ioctl=g35_fo_ioctl,
#endif //HAVE_UNLOCKED_IOCTL || LINUX_VERSION_CODE>=KERNEL_VERSION(5,0,0)

#if defined(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL)
    .compat_ioctl=g35_fo_ioctl_compat,
#endif //defined(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL)
    .mmap=g35_fo_mmap,
};

#if LINUX_VERSION_CODE<KERNEL_VERSION(4,15,0)
static void g35_wd_timer(unsigned long arg)
{
 struct g35ddc *dev=(struct g35ddc*)arg;
#else
static void g35_wd_timer(struct timer_list *t)
{
 struct g35ddc *dev=from_timer(dev,t,wd_timer);
#endif

    queue_work(dev->workqueue,&dev->wd_work);
}

static void g35_wd_work_handler(struct work_struct *work)
{
 struct g35ddc *dev;
 int status;

    dev=container_of(work,struct g35ddc,wd_work);

    G35_DEV_LOCK;

    if(likely(dev->state.open))
    {
        if(dev->state.power)
        {
            status=dev->func_wd(dev);

            if(likely(status==0))
            {
                dev->wd_counter=0;
            }
            else if(status>0)
            {
                dev->wd_counter++;

                if(unlikely(dev->wd_counter>=24))
                {
                    dev->state.temperature=1;
                    g35_set_power(dev,0,0);
                    dev->wd_counter=0;
                }
            }
        }

        dev->wd_timer.expires=jiffies+2*HZ;
        add_timer(&dev->wd_timer);
    }

    G35_DEV_UNLOCK;
}

static void g35_initialize(struct g35ddc *dev)
{
 int i;

    dev->state.connected=1;
    g35_mutex_init(&dev->mutex);
    init_waitqueue_head(&dev->power_wait_queue);


    for(i=0;i<G35_IO_STREAM_COUNT;i++)
    {
        init_rwsem(&dev->io_stream[i].rw_semaphore);

        INIT_LIST_HEAD(&dev->io_stream[i].complete_buffers.list);
        sema_init(&dev->io_stream[i].complete_buffers.semaphore,0);
        spin_lock_init(&dev->io_stream[i].complete_buffers.spinlock);

        INIT_LIST_HEAD(&dev->io_stream[i].pending_buffers.list);
        spin_lock_init(&dev->io_stream[i].pending_buffers.spinlock);
    }

    INIT_WORK(&dev->wd_work,g35_wd_work_handler);

#if LINUX_VERSION_CODE<KERNEL_VERSION(4,15,0)
    init_timer(&dev->wd_timer);

    dev->wd_timer.function=g35_wd_timer;
    dev->wd_timer.data=(unsigned long)dev;
#else
    timer_setup(&dev->wd_timer,g35_wd_timer,0);
#endif
}

static void g35_uninitialize(struct g35ddc *dev)
{
    g35_mutex_destroy(&dev->mutex);
}

static void g35_delete(struct g35ddc *dev)
{
    g35_set_power(dev,0,1);

    dev->func_delete(dev);

    g35_uninitialize(dev);

    kfree(dev);
}

static void g35_release_callback(struct device *dev)
{
 struct g35ddc *g35ddc;

    g35ddc=container_of(dev,struct g35ddc,dev);

    g35_delete(g35ddc);
}

static int g35_register(struct g35ddc *dev)
{
 int status;
 int index;
 dev_t dev_no;

    G35_REG_LOCK;

    for(index=0;index<G35_MAX_DEVICE_COUNT;index++)
    {
        if(g35_registered_devices[index]==NULL)
        {
            break;
        }
    }

    if(index>=G35_MAX_DEVICE_COUNT)
    {
        G35_REG_UNLOCK;

        G35_ERROR("no free device slot available!\n");
        return -ENFILE;
    }

    dev->index=index;

    dev_no=MKDEV(MAJOR(g35_dev0),index);

    cdev_init(&dev->cdev,&g35_file_operations);

    dev->cdev.owner=THIS_MODULE;

    status=cdev_add(&dev->cdev,dev_no,1);

    if(unlikely(status<0))
    {
        G35_REG_UNLOCK;
        G35_ERROR("char device registration failed!\n");
        return status;
    }

    memset(&dev->dev,0,sizeof(dev->dev));

    dev->dev.devt=dev_no;
    dev->dev.class=&g35_class;
    dev->dev.release=g35_release_callback;

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,26)
    dev_set_name(&dev->dev,G35_DEVICE_NAME"%d",index);
#else
    snprintf(dev->dev.bus_id,sizeof(dev->dev.bus_id),G35_DEVICE_NAME"%d",index);
#endif

    status=device_register(&dev->dev);

    if(unlikely(status<0))
    {
        cdev_del(&dev->cdev);
        G35_REG_UNLOCK;
        G35_ERROR("failed to register device!\n");
        return status;
    }

    g35_registered_devices[index]=dev;

    G35_REG_UNLOCK;

    return 0;
}

static void g35_unregister(struct g35ddc *dev)
{
    G35_REG_LOCK;

    if(likely(g35_registered_devices[dev->index]==dev))
    {
        g35_registered_devices[dev->index]=NULL;

        G35_REG_UNLOCK;

        cdev_del(&dev->cdev);
        device_unregister(&dev->dev);
    }
    else
    {
        G35_REG_UNLOCK;
    }
}

static int g35_set_power(struct g35ddc *dev,uint8_t power,uint8_t lock)
{
 int status;
 int i;

    power=power?1:0;

    if(lock)
    {
        G35_DEV_LOCK;
    }

    if(dev->state.power==power)
    {
        status=0;
    }
    else
    {
        if(power)
        {
            dev->state.suspended=0;
            dev->state.temperature=0;
        }
        else
        {
            dev->func_if_stop(dev);
            dev->func_dac_stop(dev);

            for(i=0;i<G35_DDC1_COUNT;i++)
            {
                dev->func_ddc1_stop(dev,i);
            }
            dev->func_ddc1_playback_stop(dev);
        }

        status=dev->func_set_power(dev,power);

        if(likely(status>=0))
        {
            if(power)
            {
                dev->state.power=1;
                msleep(10);
            }
            else
            {
                dev->state.power=0;
                dev->state.fpga=0;

                wake_up_all(&dev->power_wait_queue);
            }
        }
    }

    if(lock)
    {
        G35_DEV_UNLOCK;
    }

    return status;
}

static int g35_configure_fpga(struct g35ddc *dev,void __user *user_buffer,uint32_t size)
{
 int status;

    if(unlikely(size==0 || user_buffer==NULL))
    {
        return -EINVAL;
    }

    G35_DEV_LOCK;

    if(unlikely(!dev->state.power))
    {
        G35_DEV_UNLOCK;
        return -EPERM;
    }

    status=dev->func_configure_fpga(dev,user_buffer,size);

    if(likely(status>=0))
    {
        dev->state.fpga=1;
        msleep(50);
    }

    G35_DEV_UNLOCK;

    return status;
}

static int g35_rf_control1(struct g35ddc *dev,uint32_t value)
{
 int status;

    G35_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        status=dev->func_rf_control1(dev,value);
    }
    else
    {
        status=-EPERM;
    }

    G35_DEV_UNLOCK;

    return status;
}

static int g35_rf_control2(struct g35ddc *dev,uint32_t value)
{
 int status;

    G35_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        status=dev->func_rf_control2(dev,value);
    }
    else
    {
        status=-EPERM;
    }

    G35_DEV_UNLOCK;

    return status;
}

static int g35_set_noise_blanker(struct g35ddc *dev,uint16_t value)
{
 int status;

    G35_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        status=dev->func_set_noise_blanker(dev,value);
    }
    else
    {
        status=-EPERM;
    }

    G35_DEV_UNLOCK;

    return status;
}

static int g35_if_start(struct g35ddc *dev,uint16_t period)
{
 int status;

    if(unlikely(period==0))
    {
        return -EINVAL;
    }

    G35_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        if(unlikely(dev->io_stream[G35_IO_STREAM_IN_IF].active ))
        {
            status=-EBUSY;
        }
        else
        {
            status=dev->func_if_start(dev,period);
        }
    }
    else
    {
        status=-EPERM;
    }

    G35_DEV_UNLOCK;

    return status;
}

static void g35_if_stop(struct g35ddc *dev)
{
    G35_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        dev->func_if_stop(dev);
    }

    G35_DEV_UNLOCK;
}


inline static int g35_get_complete_buffer(struct g35ddc *dev,uint32_t stream_index,unsigned long arg)
{
 uintptr_t user_buffer;
 int status;

    if(unlikely(stream_index>=G35_IO_STREAM_COUNT))
    {
        return -EINVAL;
    }

    status=g35_io_stream_get_complete_buffer(dev,&dev->io_stream[stream_index],&user_buffer);

    if(unlikely(status<0))
    {
        return status;
    }

    return __put_user(user_buffer,(uintptr_t __user*)arg);
}

#if defined(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL)

inline static int g35_get_complete_buffer_compat(struct g35ddc *dev,uint32_t stream_index,unsigned long arg)
{
 uintptr_t user_buffer;
 int status;

    status=g35_io_stream_get_complete_buffer(dev,&dev->io_stream[stream_index],&user_buffer);

    if(unlikely(status<0))
    {
        return status;
    }

    return __put_user(ptr_to_compat((void __user*)user_buffer),(compat_uptr_t __user*)compat_ptr(arg));
}

#endif //defined(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL)


static int g35_put_buffer(struct g35ddc *dev,uint32_t stream_index,void __user *user_buffer,int write)
{
 struct g35_io_buffer *io_buffer;
 struct g35_io_stream *io_stream;
 int status;

    io_stream=dev->io_stream+stream_index;

    down_read(&io_stream->rw_semaphore);

    if(unlikely(!io_stream->active))
    {
        up_read(&io_stream->rw_semaphore);
        return -EPERM;
    }

    status=g35_io_buffer_prepare((uintptr_t)user_buffer,io_stream->buffer_size,write,&io_buffer);

    if(unlikely(status))
    {
        up_read(&io_stream->rw_semaphore);
        return status;
    }

    status=dev->func_put_buffer(dev,stream_index,io_buffer);

    up_read(&io_stream->rw_semaphore);

    if(unlikely(status<0))
    {
        g35_io_buffer_free(io_buffer);
    }

    return status;
}

static int g35_get_adc_level(struct g35ddc *dev,uint16_t *level)
{
 int status;

    G35_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        status=dev->func_get_adc_level(dev,level);
    }
    else
    {
        status=-EPERM;
    }

    G35_DEV_UNLOCK;

    return status;
}

static int g35_ddc1_prepare(struct g35ddc *dev,uint32_t channel,struct g35_ddc1_params *params,uint32_t __user *user_coeffs,uint32_t num_coeffs,uint32_t buffer_size,uint32_t bytes_per_sec)
{
 int status;
 uint32_t *coeffs;
 struct g35_io_stream *io_stream;

    if(unlikely(buffer_size>G35_MAX_BUFFER_SIZE))
    {
        return -E2BIG;
    }

    coeffs=kmalloc(num_coeffs*sizeof(uint32_t),GFP_KERNEL);

    if(unlikely(coeffs==NULL))
    {
        return -ENOMEM;
    }

    if(unlikely(copy_from_user(coeffs,user_coeffs,num_coeffs*sizeof(uint32_t))!=0))
    {
        kfree(coeffs);
        return -EFAULT;
    }

    G35_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        io_stream=dev->io_stream+G35_IO_STREAM_IN_DDC+channel;

        if(unlikely(io_stream->active))
        {
            status=-EBUSY;
        }
        else
        {
            io_stream->buffer_size=buffer_size;
            io_stream->bytes_per_sec=bytes_per_sec;

            dev->ddc1_params[channel]=*params;

            status=dev->func_ddc1_prepare(dev,channel,coeffs,num_coeffs);
        }
    }
    else
    {
        status=-EPERM;
    }

    G35_DEV_UNLOCK;

    kfree(coeffs);

    return status;
}

static int g35_ddc1_start(struct g35ddc *dev,uint32_t channel)
{
 int status;

    G35_DEV_LOCK;

    if(dev->state.fpga)
    {
        status=dev->func_ddc1_start(dev,channel);
    }
    else
    {
        status=-EPERM;
    }

    G35_DEV_UNLOCK;

    return status;
}

static void g35_ddc1_stop(struct g35ddc *dev,uint32_t channel)
{
    G35_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        dev->func_ddc1_stop(dev,channel);
    }

    G35_DEV_UNLOCK;
}


static int g35_ddc_playback_set_frequency(struct g35ddc *dev,uint32_t tuning_word)
{
 int status;

    G35_DEV_LOCK;

    if(likely(dev->state.fpga && dev->io_stream[G35_IO_STREAM_IN_DDC_PLAYBACK].active))
    {
        dev->ddc1_params[G35_DDC_WIDE].tuning_word=tuning_word;
        status=dev->func_ddc1_update_frequency(dev,G35_DDC_WIDE);
    }
    else
    {
        status=-EPERM;
    }

    G35_DEV_UNLOCK;

    return status;
}

static int g35_ddc1_set_frequency(struct g35ddc *dev,uint32_t channel,uint32_t tuning_word)
{
 int status;

    G35_DEV_LOCK;

    if(likely(dev->state.fpga && dev->io_stream[G35_IO_STREAM_IN_DDC+channel].active))
    {
        dev->ddc1_params[channel].tuning_word=tuning_word;

        status=dev->func_ddc1_update_frequency(dev,channel);
    }
    else
    {
        status=-EPERM;
    }

    G35_DEV_UNLOCK;

    return status;
}

static int g35_ddc1_set_phase_shift(struct g35ddc *dev,uint32_t channel,uint32_t tuning_word)
{
 int status;

    G35_DEV_LOCK;

    if(likely(dev->state.fpga && dev->io_stream[G35_IO_STREAM_IN_DDC+channel].active))
    {
        dev->ddc1_params[channel].phase_shift=tuning_word;

        status=dev->func_ddc1_update_phase_shift(dev,channel);
    }
    else
    {
        status=-EPERM;
    }

    G35_DEV_UNLOCK;

    return status;
}


static int g35_ddc1_playback_start(struct g35ddc *dev,struct g35_ddc1_params *params,uint32_t __user *user_coeffs,uint32_t buffer_size,uint32_t bytes_per_sec)
{
 int status;
 uint32_t *coeffs;
 uint32_t i;

    if(unlikely(buffer_size>G35_MAX_BUFFER_SIZE)) return -E2BIG;

    coeffs=kmalloc(params->fir_len*sizeof(uint32_t)/2,GFP_KERNEL);

    if(unlikely(coeffs==NULL)) return -ENOMEM;

    if(unlikely(copy_from_user(coeffs,user_coeffs,params->fir_len*sizeof(uint32_t)/2)!=0)) {kfree(coeffs);return -EFAULT;}

    G35_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        if(unlikely(dev->io_stream[G35_IO_STREAM_IN_DDC_PLAYBACK].active))
        {
            status=-EBUSY;
        }
        else
        {
            status=0;

            for(i=0;i<G35_DDC_COUNT;i++)
            {
                if(dev->io_stream[G35_IO_STREAM_IN_DDC+i].active) {status=-EBUSY;break;}
            }
        }

        if (status==0)
        {
            dev->io_stream[G35_IO_STREAM_OUT_DDC_PLAYBACK].bytes_per_sec=0;
            dev->io_stream[G35_IO_STREAM_IN_DDC_PLAYBACK].bytes_per_sec=bytes_per_sec;
            dev->io_stream[G35_IO_STREAM_OUT_DDC_PLAYBACK].buffer_size=buffer_size;

            dev->ddc1_params[G35_DDC_WIDE]= *params;
            dev->ddc1_params[G35_DDC_WIDE].control_word|=0x8000;//playback mode

            status=dev->func_ddc1_playback_start(dev,coeffs,params->fir_len);
        }
    }
    else
    {
        status=-EPERM;
    }

    G35_DEV_UNLOCK;
    kfree(coeffs);
    return status;
}

static void g35_ddc1_playback_stop(struct g35ddc *dev)
{
    G35_DEV_LOCK;

    if(dev->state.fpga)
    {
        dev->func_ddc1_playback_stop(dev);
    }

    G35_DEV_UNLOCK;
}

static int g35_dac_start(struct g35ddc *dev,struct g35_dac_params *params)
{
 int status;

    if(unlikely(params->dac_control==0 || params->bytes_per_sec==0))
    {
        return -EINVAL;
    }

    G35_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        if(unlikely(dev->io_stream[G35_IO_STREAM_OUT_DAC].active ))
        {
            status=-EBUSY;
        }
        else
        {
            dev->io_stream[G35_IO_STREAM_OUT_DAC].bytes_per_sec=params->bytes_per_sec;
            dev->io_stream[G35_IO_STREAM_OUT_DAC].buffer_size=params->buffer_size;

            status=dev->func_dac_start(dev,params->dac_control);
        }
    }
    else
    {
        status=-EPERM;
    }

    G35_DEV_UNLOCK;

    return status;
}

static void g35_dac_stop(struct g35ddc *dev)
{
    G35_DEV_LOCK;

    if(dev->state.fpga)
    {
        dev->func_dac_stop(dev);
    }

    G35_DEV_UNLOCK;
}

static void g35_cancel_buffers(struct g35ddc *dev,int stream_index)
{
 struct g35_io_stream *io_stream;

    io_stream=&dev->io_stream[stream_index];

    down_read(&io_stream->rw_semaphore);

    if(io_stream->active)
    {
        dev->func_cancel_buffers(dev,stream_index);
    }

    up_read(&io_stream->rw_semaphore);
}

static int g35_wait_for_power_off(struct g35ddc *dev)
{
 DECLARE_WAITQUEUE(wait_queue_entry,current);

    add_wait_queue(&dev->power_wait_queue,&wait_queue_entry);

    for(;;)
    {
        G35_DEV_LOCK;

        if(!dev->state.power || !dev->state.open || !dev->state.connected)
        {
            G35_DEV_UNLOCK;
            break;
        }

        G35_DEV_UNLOCK;

        set_current_state(TASK_INTERRUPTIBLE);

        schedule();

        set_current_state(TASK_RUNNING);

        if(signal_pending(current))
        {
            remove_wait_queue(&dev->power_wait_queue,&wait_queue_entry);
            return -EINTR;
        }
    }

    remove_wait_queue(&dev->power_wait_queue,&wait_queue_entry);
    return 0;
}


static int g35_get_device_state(struct g35ddc *dev,void __user *user_buffer)
{
 int status;

    G35_DEV_LOCK;

    status=copy_to_user(user_buffer,&dev->state,sizeof(dev->state))?-EFAULT:0;

    G35_DEV_UNLOCK;

    return status;
}


static int g35_coh_arm(struct g35ddc *dev,void __user *user_buffer)
{
 int status;
 uint8_t master;

    if(unlikely(get_user(master,(uint8_t __user*)user_buffer)<0))
    {
        return -EFAULT;
    }

    G35_DEV_LOCK;

    if(unlikely(dev->config[2]<0x35 || !dev->state.fpga))
    {
        status=-EPERM;
    }
    else
    {
        status=dev->func_coh_arm(dev,master);
    }

    G35_DEV_UNLOCK;

    return status;
}


static int g35_coh_trigger(struct g35ddc *dev)
{
 int status;

    G35_DEV_LOCK;

    if(unlikely(dev->config[2]<0x35 || !dev->state.fpga))
    {
        status=-EPERM;
    }
    else
    {
        status=dev->func_coh_trigger(dev);
    }

    G35_DEV_UNLOCK;

    return status;
}

static int g35_coh_set(struct g35ddc *dev,void __user *user_buffer)
{
 int status;
 uint8_t value;

    if(unlikely(get_user(value,(uint8_t __user*)user_buffer)<0))
    {
        return -EFAULT;
    }

    G35_DEV_LOCK;

    if(unlikely(dev->config[2]<0x35 || dev->state.power))
    {
        status=-EPERM;
    }
    else
    {
        status=dev->func_coh_set(dev,value);
    }

    G35_DEV_UNLOCK;

    return status;
}


static int g35_coh_get(struct g35ddc *dev,void __user *user_buffer)
{
 int status;
 uint8_t value;

    G35_DEV_LOCK;

    if(unlikely(dev->config[2]<0x35 || dev->state.power))
    {
        status=-EPERM;
    }
    else
    {
        status=dev->func_coh_get(dev,&value);

        if(status==0)
        {
            status=put_user(value,(uint8_t __user*)user_buffer);
        }
    }

    G35_DEV_UNLOCK;

    return status;
}

static int g35_ddc1_get_timestamp(struct g35ddc *dev,uint32_t channel,void __user *user_buffer)
{
 union g35_timestamp timestamp;
 int status;

    G35_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        status=dev->func_ddc1_get_timestamp(dev,channel,&timestamp);

        if(likely(status==0))
        {
            status=copy_to_user(user_buffer,&timestamp,sizeof(timestamp))?-EFAULT:0;
        }
    }
    else
    {
        status=-EPERM;
    }

    G35_DEV_UNLOCK;

    return status;
}
#ifdef __G35_PCI__

static int g35_pci_initialize(struct g35ddc_pci *dev);


static void g35_pci_setup(struct g35ddc_pci *dev)
{
 uint32_t i,j;

    bar4_write32(GPIO_INT_MASK_SET,0xFFFFFFFF);

    bar4_write32(INT_CFG0,0);
    bar4_write32(INT_CFG1,0);
    bar4_write32(INT_CFG2,0);
    bar4_write32(INT_CFG3,0);

    i=bar4_read32(PCI_SYS_CFG_SYSTEM);
    i&=~0x3000;
    bar4_write32(PCI_SYS_CFG_SYSTEM,i);
    bar4_read32(PCI_SYS_CFG_SYSTEM);

    i=bar4_read32(PCI_SYS_CFG_SYSTEM);
    i&=~0xC000;
    bar4_write32(PCI_SYS_CFG_SYSTEM,i);
    bar4_read32(PCI_SYS_CFG_SYSTEM);

    if(unlikely((bar4_read32(CLK_CSR) & 0x0FF7F3FF)!=0x0001F07C))
    {
        bar4_write32(CLK_CSR,0x0001F07C | 0x8080000);
        bar4_read32(CLK_CSR);
        bar4_write32(CLK_CSR,0x0001F07C);

        DEBUG("LCLK changed");

        for(j=0;j<10;j++)
        {
            if((bar4_read32(CLK_CSR) & 0xC0000000)==0xC0000000)
            {
                break;
            }

            msleep(50);
        }
    }

    i=bar4_read32(PCI_SYS_CFG_SYSTEM);
    i|=0x4000;
    bar4_write32(PCI_SYS_CFG_SYSTEM,i);
    bar4_read32(PCI_SYS_CFG_SYSTEM);


    msleep(250);


    bar4_write32(GPIO_INT_TYPE,GPIO_BIT_DMA_IRQ);
    bar4_write32(GPIO_INT_VALUE,GPIO_BIT_DMA_IRQ);

    bar4_read32(GPIO_INT_STATUS);
    bar4_read32(INT_STAT);
}

static int g35_pci_probe(struct pci_dev *pci_dev, const struct pci_device_id *id)
{
 int status;
 struct g35ddc_pci *dev;
 unsigned long start,len;

    if(pci_dev->vendor!=G35_PCI_VENDOR || pci_dev->device!=G35_PCI_DEVICE ||
       pci_dev->subsystem_vendor!=G35_PCI_SUBSYSTEM_VENDOR || pci_dev->subsystem_device!=G35_PCI_SUBSYSTEM_DEVICE)
    {
        G35_ERROR("unsuported device!\n");
        return -ENODEV;
    }

    dev=kmalloc(sizeof(struct g35ddc_pci),GFP_KERNEL);

    if(unlikely(dev==NULL))
    {
        G35_ERROR("failed to allocate memory for device object!\n");
        return -ENOMEM;
    }

    memset(dev,0,sizeof(*dev));

    dev->pci_dev=pci_dev_get(pci_dev);
    dev->g35_dev.config[0]=G35_INTERFACE_TYPE_PCIE;

    status=pci_enable_device(dev->pci_dev);

    if(unlikely(status))
    {
        G35_ERROR("failed to enable device!\n");
        goto error_kfree;
    }

    pci_set_drvdata(dev->pci_dev,dev);

    if(dma_set_mask(&dev->pci_dev->dev,DMA_BIT_MASK(64))==0)
    {
        if(unlikely(dma_set_coherent_mask(&dev->pci_dev->dev,DMA_BIT_MASK(64))!=0))
        {
            if(unlikely(dma_set_coherent_mask(&dev->pci_dev->dev,DMA_BIT_MASK(32))!=0))
            {
                G35_ERROR("no suitable DMA available!\n");
                status=-EIO;
                goto error_pci_disable_device;
            }
        }

        dev->dma_64=1;
    }
    else
    {
        if(unlikely(dma_set_mask(&dev->pci_dev->dev,DMA_BIT_MASK(32))!=0))
        {
            G35_ERROR("no suitable DMA available!\n");
            status=-EIO;
            goto error_pci_disable_device;
        }

        if(unlikely(dma_set_coherent_mask(&dev->pci_dev->dev,DMA_BIT_MASK(32))!=0))
        {
            G35_ERROR("no suitable DMA available!\n");
            status=-EIO;
            goto error_pci_disable_device;
        }
    }


    pci_set_master(dev->pci_dev);

    status=pci_request_regions(dev->pci_dev,pci_name(dev->pci_dev));

    if(unlikely(status))
    {
        G35_ERROR("failed to reserve PCI I/O and memory resources!\n");
        goto error_pci_disable_device;
    }

    /*if(pci_enable_msi(dev->pci_dev)==0)
    {
        dev->msi=1;
    }
    else
    {
        pci_disable_msi(dev->pci_dev);
    }*/

    if(unlikely((pci_resource_flags(dev->pci_dev,0) & IORESOURCE_MEM)==0 ||
                (pci_resource_flags(dev->pci_dev,4) & IORESOURCE_MEM)==0))
    {
        G35_ERROR("device memory resources not found\n");
        status=-EIO;
        goto error_pci_release_regions;
    }

    start=pci_resource_start(dev->pci_dev,0);
    len=pci_resource_len(dev->pci_dev,0);
#if LINUX_VERSION_CODE>=KERNEL_VERSION(5,6,0)
    dev->bar0=ioremap(start,len);
#else
    dev->bar0=ioremap_nocache(start,len);
#endif

    if(unlikely(dev->bar0==NULL))
    {
        G35_ERROR("failed to map memory resource!\n");
        status=-ENOMEM;
        goto error_pci_release_regions;
    }

    start=pci_resource_start(dev->pci_dev,4);
    len=pci_resource_len(dev->pci_dev,4);

#if LINUX_VERSION_CODE>=KERNEL_VERSION(5,6,0)
    dev->bar4=ioremap(start,len);
#else
    dev->bar4=ioremap_nocache(start,len);
#endif

    if(unlikely(dev->bar4==NULL))
    {
        G35_ERROR("failed to map memory resource!\n");
        status=-ENOMEM;
        goto error_unmap_bar0;
    }

    g35_initialize(&dev->g35_dev);

    status=g35_pci_initialize(dev);

    if(unlikely(status))
    {
        g35_uninitialize(&dev->g35_dev);
        goto error_unmap_bar4;
    }

    status=g35_register(&dev->g35_dev);

    if(unlikely(status<0))
    {
        g35_delete(&dev->g35_dev);
        return status;
    }

    return 0;

    error_unmap_bar4:
        iounmap(dev->bar4);

    error_unmap_bar0:
        iounmap(dev->bar0);

    error_pci_release_regions:
        if(dev->msi)
        {
            pci_disable_msi(dev->pci_dev);
        }

        pci_release_regions(dev->pci_dev);

    error_pci_disable_device:
        pci_set_drvdata(dev->pci_dev,NULL);
        pci_disable_device(dev->pci_dev);

    error_kfree:
        pci_dev_put(dev->pci_dev);

        kfree(dev);

    return status;
}

static void g35_pci_remove(struct pci_dev *pci_dev)
{
 struct g35ddc_pci *dev;

    dev=pci_get_drvdata(pci_dev);

    if(likely(dev!=NULL))
    {
        pci_set_drvdata(pci_dev,NULL);

        dev->g35_dev.state.connected=0;

        g35_unregister(&dev->g35_dev);
    }
}


static void g35_pci_delete(struct g35ddc_pci *dev)
{
 int i;

    iounmap(dev->bar4);
    iounmap(dev->bar0);

    if(dev->msi)
    {
        pci_disable_msi(dev->pci_dev);
    }

    pci_disable_device(dev->pci_dev);
    pci_release_regions(dev->pci_dev);

    pci_dev_put(dev->pci_dev);

    for(i=0;i<G35_IO_STREAM_COUNT;i++)
    {
        g35_mutex_destroy(&dev->dma_channel[i].mutex);
    }
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
irqreturn_t g35_pci_interrupt_handler(int irqno, void *dev_id)
#else
irqreturn_t g35_pci_interrupt_handler(int irqno, void *dev_id, struct pt_regs *regs)
#endif
{
 struct g35ddc_pci *dev;
 unsigned long flags;
 uint32_t int_stat;
 uint32_t gpio_int_stat;

    dev=dev_id;

    spin_lock_irqsave(&dev->irq_spinlock,flags);

    int_stat=bar4_read32(INT_STAT);

    if((int_stat & bar4_read32(INT_CFG0))==0)
    {
        spin_unlock_irqrestore(&dev->irq_spinlock,flags);
        return IRQ_NONE;
    }

    if(unlikely(int_stat & 1))
    {
        bar4_read32(TWI_IRT_STATUS);
    }

    if(unlikely(int_stat & 2))
    {
        bar4_read32(FCL_IRQ);
    }

    if(likely(int_stat & IRQ_GPIO))
    {
        gpio_int_stat=bar4_read32(GPIO_INT_STATUS);

        if(likely(dev->g35_dev.state.fpga && (gpio_int_stat & GPIO_BIT_DMA_IRQ)))
        {
            dev->dma_status_out|=bar0_read32(REG_DMA_IRQ_STATUS_OUT);
            dev->dma_status_in|=bar0_read32(REG_DMA_IRQ_STATUS_IN);

            queue_work(dev->g35_dev.workqueue,&dev->irq_work);
        }
    }

    bar4_write32(INT_STAT,0xFFFFFFF3);
    bar4_read32(INT_STAT);

    spin_unlock_irqrestore(&dev->irq_spinlock,flags);

    return IRQ_HANDLED;
}

static void g35_pci_free_sg_list_memory(struct g35ddc_pci *dev)
{
 uint32_t i,j;

    for(i=0;i<G35_IO_STREAM_COUNT;i++)
    {
        for(j=0;j<2;j++)
        {
            if(likely(dev->dma_channel[i].subchannel[j].sg_mem_cpu!=NULL))
            {
                dma_free_coherent(&dev->pci_dev->dev,dev->dma_sg_list_memory_size+PAGE_SIZE-1,
                    dev->dma_channel[i].subchannel[j].sg_mem_cpu,
                    dev->dma_channel[i].subchannel[j].sg_mem);

                dev->dma_channel[i].subchannel[j].sg_mem_cpu=NULL;
                dev->dma_channel[i].subchannel[j].sg_mem=0;
                dev->dma_channel[i].subchannel[j].aligned_sg_mem_cpu=NULL;
                dev->dma_channel[i].subchannel[j].aligned_sg_mem=0;
            }
        }
    }
}

static int g35_pci_allocate_sg_list_memory(struct g35ddc_pci *dev)
{
 uint32_t i,j;
 void *addr;
 dma_addr_t dma_addr;

    for(i=0;i<G35_IO_STREAM_COUNT;i++)
    {
        for(j=0;j<2;j++)
        {
            addr=dma_alloc_coherent(&dev->pci_dev->dev,dev->dma_sg_list_memory_size+PAGE_SIZE-1,&dma_addr,GFP_ATOMIC);

            if(unlikely(addr==NULL))
            {
                g35_pci_free_sg_list_memory(dev);
                return -ENOMEM;
            }

            dev->dma_channel[i].subchannel[j].sg_mem_cpu=addr;
            dev->dma_channel[i].subchannel[j].sg_mem=dma_addr;

            dev->dma_channel[i].subchannel[j].aligned_sg_mem_cpu=(void*)(((uintptr_t)addr+PAGE_SIZE-1) & PAGE_MASK);
            dev->dma_channel[i].subchannel[j].aligned_sg_mem=(dma_addr+PAGE_SIZE-1) & PAGE_MASK;
        }
    }

    return 0;
}

static int g35_pci_open(struct g35ddc_pci *dev)
{
 int status;

    status=g35_pci_allocate_sg_list_memory(dev);

    if(unlikely(status))
    {
        G35_ERROR("failed to allocate coherent DMA memory!\n");
        return status;
    }

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,22)
    status=request_irq(dev->pci_dev->irq,g35_pci_interrupt_handler,IRQF_SHARED,G35_DEVICE_NAME,dev);
#else
    status=request_irq(dev->pci_dev->irq,g35_pci_interrupt_handler,SA_SHIRQ|SA_INTERRUPT,G35_DEVICE_NAME,dev);
#endif

    if(unlikely(status))
    {
        G35_ERROR("failed to register interrupt handler!\n");
        g35_pci_free_sg_list_memory(dev);
        return status;
    }

    return 0;
}

static void g35_pci_close(struct g35ddc_pci *dev)
{
    free_irq(dev->pci_dev->irq,dev);
    g35_pci_free_sg_list_memory(dev);
}

static int g35_pci_set_power(struct g35ddc_pci *dev,uint32_t power)
{
 unsigned long flags;
 uint32_t u32;

    if(power)
    {
        if(unlikely(dev->disabled))
        {
            return -EIO;
        }

        gpio_set(GPIO_BIT_FE_POWER);

        if(likely(dev->g35_dev.config[1]))
        {
            msleep(1);

            if((bar4_read32(GPIO_INPUT_VALUE) & GPIO_BIT_POWER_STATE)==0)
            {
                gpio_clr(GPIO_BIT_FE_POWER);
                return -EIO;
            }
        }
    }
    else
    {
        bar4_write32(GPIO_INT_MASK_SET,GPIO_BIT_DMA_IRQ);

        u32=bar4_read32(PCI_SYS_CFG_SYSTEM);
        u32&=~(0xC000);
        bar4_write32(PCI_SYS_CFG_SYSTEM,u32);
        bar4_read32(PCI_SYS_CFG_SYSTEM);//Read back to ensure write completion

        gpio_clr(GPIO_BIT_FPGA_RESET);  //FPGA reset

        //Disable local bus
        u32&=~(0x3000);
        bar4_write32(PCI_SYS_CFG_SYSTEM,u32);
        bar4_read32(PCI_SYS_CFG_SYSTEM);//Read back to ensure write completion

        bar4_write32(FCL_IODATA_OUT,0x00);
        msleep(1);
        bar4_write32(FCL_IODATA_OUT,0x40);
        msleep(1);

        spin_lock_irqsave(&dev->irq_spinlock,flags);

        bar4_write32(INT_CFG0,0);
        bar4_read32(INT_CFG0);

        spin_unlock_irqrestore(&dev->irq_spinlock,flags);

        gpio_clr(GPIO_BIT_FE_POWER);
    }

    return 0;
}

static int g35_pci_configure_fpga(struct g35ddc_pci *dev,void __user *user_buffer,uint32_t size)
{
 uint8_t *buffer;
 uint8_t u8;
 uint32_t u32;
 uint32_t i,j;
 unsigned long flags;

    buffer=vmalloc(size);

    if(buffer==NULL)
    {
        return -ENOMEM;
    }

    if(unlikely(copy_from_user(buffer,user_buffer,size)!=0))
    {
        vfree(buffer);
        return -EFAULT;
    }


    u32=bar4_read32(PCI_SYS_CFG_SYSTEM);

    u32&=~(0x3000);
    bar4_write32(PCI_SYS_CFG_SYSTEM,u32);
    bar4_read32(PCI_SYS_CFG_SYSTEM);

    gpio_clr(GPIO_BIT_FPGA_RESET);

    msleep(10);

    bar4_write32(FCL_CLK_DIV,0);

    bar4_write32(FCL_CTRL,0x40);
    u32=bar4_read32(FCL_CTRL);
    if(unlikely(u32!=0x40))
    {
        goto error_fcl;
    }

    bar4_write32(FCL_CTRL,0x00);
    bar4_read32(FCL_IRQ);

    bar4_write32(FCL_EN,0x17);

    bar4_write32(FCL_IODATA_OUT,0x04);
    bar4_write32(FCL_IODATA_OUT,0x00);

	for(i=0;i<1000000;i++)
	{
        if((bar4_read32(FCL_STATUS) & 0x28)==0)
        {
            break;
        }
	}

    if(unlikely(i>=1000000))
    {
        goto error_fcl;
    }

    bar4_write32(FCL_IODATA_OUT,0x04);

    udelay(50);

    for(i=0;i<1000000;i++)
    {
        if((bar4_read32(FCL_STATUS) & 0x20)!=0)
        {
            break;
        }
	}

    if(unlikely(i>=1000000))
    {
        goto error_fcl;
    }

    udelay(50);

	for(i=0;i<size;i++)
	{
        u8=buffer[i];

		for(j=0;j<8;j++,u8>>=1)
		{
			if(u8&1)
			{
				bar4_write32(FCL_IODATA_OUT,0x06);
                bar4_write32(FCL_IODATA_OUT,0x07);
			}
			else
			{
                bar4_write32(FCL_IODATA_OUT,0x04);
                bar4_write32(FCL_IODATA_OUT,0x05);
			}
		}

		if((i & 0x7F) == 0)
		{
            u32=bar4_read32(FCL_STATUS);

			if(unlikely((u32 & 0x28)==0))
			{
			    goto error_fcl;
			}
		}
	}

	for(i=0;i<16;i++)
	{
        bar4_write32(FCL_IODATA_OUT,0x04);
        bar4_write32(FCL_IODATA_OUT,0x05);
	}

    bar4_write32(FCL_IODATA_OUT,0x04);

    u32=bar4_read32(FCL_STATUS);

	if(unlikely((u32 & 0x8)==0))
	{
	    goto error_fcl;
    }

    u32=bar4_read32(PCI_SYS_CFG_SYSTEM);
    u32|=0x1000;
    bar4_write32(PCI_SYS_CFG_SYSTEM,u32);
    bar4_read32(PCI_SYS_CFG_SYSTEM);

    msleep(50);

/*
    u32|=0x4000;
    bar4_write32(PCI_SYS_CFG_SYSTEM,u32);
    bar4_read32(PCI_SYS_CFG_SYSTEM);
*/
    gpio_set(GPIO_BIT_FPGA_RESET);

    msleep(80);

    bar0_write32(REG_DMA_CONFIG,dev->dma_64?1:0);

    bar4_read32(GPIO_INT_STATUS);

    spin_lock_irqsave(&dev->irq_spinlock,flags);

    bar4_write32(INT_STAT,IRQ_GPIO);
    bar4_write32(INT_CFG0,bar4_read32(INT_CFG0) | IRQ_GPIO);
    bar4_read32(INT_CFG0);

    bar4_write32(GPIO_INT_MASK_CLR,GPIO_BIT_DMA_IRQ);

    spin_unlock_irqrestore(&dev->irq_spinlock,flags);

    vfree(buffer);

    return 0;

    error_fcl:
        gpio_clr(GPIO_BIT_FPGA_RESET);

        u32|=0x4000;
        bar4_write32(PCI_SYS_CFG_SYSTEM,u32);
        bar4_read32(PCI_SYS_CFG_SYSTEM);

        vfree(buffer);

    return -EIO;
}

static int g35_pci_rf_control1(struct g35ddc_pci *dev,uint32_t value)
{
    bar0_write32(REG_RF_CONTROL1,value);
    bar0_read32(REG_RF_CONTROL1);
    return 0;
}
static int g35_pci_rf_control2(struct g35ddc_pci *dev,uint32_t value)
{
    bar0_write32(REG_RF_CONTROL2,value);
    bar0_read32(REG_RF_CONTROL2);
    return 0;
}


static int g35_pci_set_noise_blanker(struct g35ddc_pci *dev,uint16_t value)
{
    bar0_write32(REG_DATA_LIMIT,value);
    return 0;
}


static int g35_pci_mmap(struct g35ddc_pci *dev,struct vm_area_struct *vma)
{
 int status;
 unsigned long size;

    size=vma->vm_end-vma->vm_start;

    if(unlikely(vma->vm_pgoff!=0 || size!=pci_resource_len(dev->pci_dev,4)))
    {
        return -EINVAL;
    }

    vma->vm_flags|=VM_IO;
    vma->vm_page_prot=pgprot_noncached(vma->vm_page_prot);

    status=io_remap_pfn_range(vma,vma->vm_start,(pci_resource_start(dev->pci_dev,4))>>PAGE_SHIFT,
				       size,vma->vm_page_prot);

    if(unlikely(status))
    {
        return -EAGAIN;
    }

    return 0;
}

static void g35_pci_start_dma_transfer(struct g35ddc_pci *dev,struct g35_io_buffer *io_buffer,struct g35_dma_channel *channel,uint32_t subchannel_index)
{
 struct scatterlist *sg;
 int i,j;
 struct g35_dma_subchannel *subchannel;
 struct g35_sg_segment_32 *sg32;
 struct g35_sg_segment_64 *sg64;
 uint32_t reg;

    if(channel->out)
    {
        reg=REG_DMA_PTR_OUT_BASE+channel->index*4*4+subchannel_index*4*2;
    }
    else
    {
        reg=REG_DMA_PTR_IN_BASE+channel->index*4*4+subchannel_index*4*2;

        io_buffer->changed=1;
    }

    g35_io_buffer_sg_sync_for_device(io_buffer);

    subchannel=channel->subchannel+subchannel_index;

    subchannel->io_buffer=io_buffer;

    if(dev->dma_64)
    {
        sg64=subchannel->aligned_sg_mem_cpu;

        j=0;


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
        for_each_sg(io_buffer->sg_list,sg,io_buffer->sg_segment_count,i)
        {
            sg64[j].address=sg_dma_address(sg);
            sg64[j].length=sg_dma_len(sg)>>2;

            j++;

            if(j>=G35_DMA_PAGE_SIZE/sizeof(struct g35_sg_segment_64))
            {
                j=0;
                sg64=(struct g35_sg_segment_64*)((uintptr_t)sg64+G35_DMA_PAGE_SIZE);
            }
        }
#else
        sg=io_buffer->sg_list;

        for(i=0;i<io_buffer->sg_segment_count;i++)
        {
            sg64[j].address=sg_dma_address(sg);
            sg64[j].length=sg_dma_len(sg)>>2;

            j++;

            if(j>=G35_DMA_PAGE_SIZE/sizeof(struct g35_sg_segment_64))
            {
                j=0;
                sg64=(struct g35_sg_segment_64*)((uintptr_t)sg64+G35_DMA_PAGE_SIZE);
            }

            sg++;
        }
#endif

        sg64[j].address=0;
        sg64[j].length=0;

        bar0_write32(reg+4,((uint64_t)subchannel->aligned_sg_mem)>>32);
        bar0_write32(reg,((uint64_t)subchannel->aligned_sg_mem)&0xFFFFFFFF);
    }
    else
    {
        sg32=subchannel->aligned_sg_mem_cpu;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
        for_each_sg(io_buffer->sg_list,sg,io_buffer->sg_segment_count,i)
        {
            sg32[i].address=sg_dma_address(sg);
            sg32[i].length=sg_dma_len(sg)>>2;
        }
#else
        sg=io_buffer->sg_list;

        for(i=0;i<io_buffer->sg_segment_count;i++)
        {
            sg32[i].address=sg_dma_address(sg);
            sg32[i].length=sg_dma_len(sg)>>2;

            sg++;
        }
#endif

        sg32[i].address=0;
        sg32[i].length=0;

        bar0_write32(reg,subchannel->aligned_sg_mem);
    }
}

static void g35_pci_disable_dma_channel(struct g35ddc_pci *dev,uint32_t stream_index)
{
 struct g35_dma_channel *channel;
 struct g35_io_stream *io_stream;
 unsigned long flags;
 int i;

    channel=dev->dma_channel+stream_index;
    io_stream=dev->g35_dev.io_stream+stream_index;

    down_write(&io_stream->rw_semaphore);

    io_stream->active=0;

    up(&io_stream->complete_buffers.semaphore);

    up_write(&io_stream->rw_semaphore);

    spin_lock_irqsave(&dev->irq_spinlock,flags);

    if(unlikely(channel->out))
    {
        dev->dma_enabled_out&=~(1<<channel->index);

        bar0_write32(REG_DMA_ENABLE_OUT,dev->dma_enabled_out);

        dev->dma_subchannel_mask_out&=~(3<<(channel->index<<1));
        dev->dma_status_out&=dev->dma_subchannel_mask_out;

        spin_unlock_irqrestore(&dev->irq_spinlock,flags);

        bar0_read32(REG_DMA_ENABLE_OUT);
    }
    else
    {
        dev->dma_enabled_in&=~(1<<channel->index);

        bar0_write32(REG_DMA_ENABLE_IN,dev->dma_enabled_in);

        dev->dma_subchannel_mask_in&=~(3<<(channel->index<<1));
        dev->dma_status_in&=dev->dma_subchannel_mask_in;

        spin_unlock_irqrestore(&dev->irq_spinlock,flags);

        bar0_read32(REG_DMA_ENABLE_IN);
    }


    g35_io_stream_free_buffers(io_stream);

    for(i=0;i<2;i++)
    {
        if(channel->subchannel[i].io_buffer)
        {
            g35_io_buffer_free(channel->subchannel[i].io_buffer);
            channel->subchannel[i].io_buffer=NULL;
        }
    }
}

static int g35_pci_enable_dma_channel(struct g35ddc_pci *dev,uint32_t stream_index)
{
 struct g35_dma_channel *channel;
 unsigned long flags;

    if(unlikely((dev->g35_dev.io_stream[stream_index].buffer_size & 0xFF)!=0))
    {
        return -EINVAL;
    }

    channel=dev->dma_channel+stream_index;

    if(unlikely(channel->out))
    {
        bar0_write32(REG_DMA_FIFO_RESET,(1<<channel->index));
    }
    else
    {
        bar0_write32(REG_DMA_FIFO_RESET,(1<<(16+channel->index)));
    }

    channel->subchannel_index=0;

    spin_lock_irqsave(&dev->irq_spinlock,flags);

    if(unlikely(channel->out))
    {
        dev->dma_enabled_out|=1<<channel->index;

        bar0_write32(REG_DMA_ENABLE_OUT,dev->dma_enabled_out);

        dev->dma_subchannel_mask_out|=3<<(channel->index<<1);

        spin_unlock_irqrestore(&dev->irq_spinlock,flags);

        bar0_read32(REG_DMA_ENABLE_OUT);
    }
    else
    {
        dev->dma_enabled_in|=1<<channel->index;

        bar0_write32(REG_DMA_ENABLE_IN,dev->dma_enabled_in);

        dev->dma_subchannel_mask_in|=3<<(channel->index<<1);

        spin_unlock_irqrestore(&dev->irq_spinlock,flags);

        bar0_read32(REG_DMA_ENABLE_IN);
    }

    dev->g35_dev.io_stream[stream_index].active=1;

    return 0;
}

static int g35_pci_finish_dma_transfer(struct g35ddc_pci *dev,struct g35_io_stream *io_stream,struct g35_dma_channel *channel)
{
 struct g35_io_buffer *io_buffer;
 uint32_t index;

    if(likely(channel->subchannel[channel->subchannel_index].io_buffer))
    {
        g35_io_buffer_sg_unmap(channel->subchannel[channel->subchannel_index].io_buffer);
        g35_io_stream_put_complete_buffer(io_stream,channel->subchannel[channel->subchannel_index].io_buffer);
        channel->subchannel[channel->subchannel_index].io_buffer=NULL;

        io_buffer=g35_io_stream_get_pending_buffer_nolock(io_stream);

        if(likely(io_buffer!=NULL))
        {
            g35_pci_start_dma_transfer(dev,io_buffer,channel,channel->subchannel_index);
        }
    }

    index=1-channel->subchannel_index;

    if(channel->subchannel[index].io_buffer)
    {
        channel->subchannel_index=index;
        return 1;
    }

    return 0;
}

static void g35_pci_irq_work_handler(struct work_struct *work)
{
 struct g35ddc_pci *dev;
 uint32_t dma_status_in;
 uint32_t dma_status_out;
 unsigned long flags;
 int i,j;
 struct g35_io_stream *io_stream;
 struct g35_dma_channel *channel;
 uint32_t dma_status_in_ooo;
 uint32_t dma_status_out_ooo;

    dev=container_of(work,struct g35ddc_pci,irq_work);

    spin_lock_irqsave(&dev->irq_spinlock,flags);

    dma_status_in=dev->dma_status_in & dev->dma_subchannel_mask_in;
    dma_status_out=dev->dma_status_out & dev->dma_subchannel_mask_out;

    dev->dma_status_in=0;
    dev->dma_status_out=0;

    spin_unlock_irqrestore(&dev->irq_spinlock,flags);

    dma_status_in_ooo=0;
    dma_status_out_ooo=0;

    for(i=0;i<G35_DMA_CHANNEL_IN_COUNT;i++,dma_status_in>>=2)
    {
        if(dma_status_in&3)
        {
            io_stream=dev->g35_dev.io_stream+G35_IO_STREAM_IN_FIRST+i;

            down_read(&io_stream->rw_semaphore);

            if(io_stream->active)
            {
                channel=dev->dma_channel+G35_IO_STREAM_IN_FIRST+i;

                g35_mutex_lock(&channel->mutex);

                for(j=0;j<2;j++)
                {
                    if(dma_status_in & (1<<channel->subchannel_index))
                    {
                        dma_status_in&=~(1<<channel->subchannel_index);

                        if(!g35_pci_finish_dma_transfer(dev,io_stream,channel))
                        {
                            break;
                        }
                    }
                    else
                    {
                        break;
                    }
                }

                g35_mutex_unlock(&channel->mutex);

                if(unlikely(dma_status_in & 3))
                {
                    dma_status_in_ooo|=(dma_status_in&3)<<(channel->index<<1);
                }
            }

            up_read(&io_stream->rw_semaphore);
        }
    }

    for(i=0;i<G35_DMA_CHANNEL_OUT_COUNT;i++,dma_status_out>>=2)
    {
        if(dma_status_out&3)
        {
            io_stream=dev->g35_dev.io_stream+G35_IO_STREAM_OUT_FIRST+i;

            down_read(&io_stream->rw_semaphore);

            if(io_stream->active)
            {
                channel=dev->dma_channel+G35_IO_STREAM_OUT_FIRST+i;

                g35_mutex_lock(&channel->mutex);

                for(j=0;j<2;j++)
                {
                    if(dma_status_out & (1<<channel->subchannel_index))
                    {
                        dma_status_out&=~(1<<channel->subchannel_index);

                        if(!g35_pci_finish_dma_transfer(dev,io_stream,channel))
                        {
                            break;
                        }
                    }
                    else
                    {
                        break;
                    }
                }

                g35_mutex_unlock(&channel->mutex);

                if(unlikely(dma_status_out & 3))
                {
                    dma_status_out_ooo|=(dma_status_out&3)<<(channel->index<<1);
                }
            }

            up_read(&io_stream->rw_semaphore);
        }
    }

    if(unlikely(dma_status_in_ooo!=0 || dma_status_out_ooo!=0))
    {
        spin_lock_irqsave(&dev->irq_spinlock,flags);

        dev->dma_status_in|=dma_status_in_ooo;
        dev->dma_status_out|=dma_status_out_ooo;

        spin_unlock_irqrestore(&dev->irq_spinlock,flags);
    }
}

static int g35_pci_if_start(struct g35ddc_pci *dev,uint16_t period)
{
 int status;

    dev->g35_dev.io_stream[G35_IO_STREAM_IN_IF].buffer_size=65536*sizeof(int16_t);
    dev->g35_dev.io_stream[G35_IO_STREAM_IN_IF].bytes_per_sec=65536*sizeof(int16_t)*1000/period;

    bar0_write32(REG_IF_BUFF_CTRL,period);

    status=g35_pci_enable_dma_channel(dev,G35_IO_STREAM_IN_IF);

    if(unlikely(status))
    {
        bar0_write32(REG_IF_BUFF_CTRL,0);
    }

    return status;
}

static void g35_pci_if_stop(struct g35ddc_pci *dev)
{
    if(likely(dev->g35_dev.io_stream[G35_IO_STREAM_IN_IF].active))
    {
        g35_pci_disable_dma_channel(dev,G35_IO_STREAM_IN_IF);

        bar0_write32(REG_IF_BUFF_CTRL,0);
    }
}

static int g35_pci_put_buffer(struct g35ddc_pci *dev,uint32_t stream_index,struct g35_io_buffer *io_buffer)
{
 int status;
 struct g35_dma_channel *channel;

    channel=dev->dma_channel+stream_index;

    status=g35_io_buffer_sg_map(io_buffer,dev->pci_dev,channel->out?DMA_TO_DEVICE:DMA_FROM_DEVICE);

    if(unlikely(status))
    {
        return status;
    }

    g35_mutex_lock(&channel->mutex);

    if(channel->subchannel[0].io_buffer==NULL)
    {
        if(channel->subchannel[1].io_buffer==NULL)
        {
            channel->subchannel_index=0;
        }

        g35_pci_start_dma_transfer(dev,io_buffer,channel,0);
    }
    else
    {
        if(channel->subchannel[1].io_buffer==NULL)
        {
            g35_pci_start_dma_transfer(dev,io_buffer,channel,1);
        }
        else
        {
            g35_io_stream_put_pending_buffer_nolock(dev->g35_dev.io_stream+stream_index,io_buffer);
        }
    }

    g35_mutex_unlock(&channel->mutex);

    return 0;
}

static int g35_pci_get_adc_level(struct g35ddc_pci *dev,uint16_t *level)
{
    *level=(uint16_t)bar0_read32(REG_ADC_LEVEL);
    return 0;
}

static void g35_pci_ddc1_setup(struct g35ddc_pci *dev,uint32_t channel,uint32_t *coeffs,uint32_t num_coeffs)
{
 struct g35_ddc1_params *params;
 uint32_t reg_offset;
 uint32_t address;
 int i;

    params=dev->g35_dev.ddc1_params+channel;

    reg_offset=channel*REG_DDC1_COUNT*4;

    bar0_write32(REG_DDC1_CTRL1+reg_offset,params->control_word & 0xFFFE);

    bar0_write32(REG_DDC1_START_PHASE+reg_offset,params->phase_shift);
    bar0_write32(REG_DDC1_TUNING_WORD+reg_offset,params->tuning_word);
    bar0_write32(REG_DDC1_CIC_CTRL1+reg_offset,params->cic1 | (((uint32_t)params->cic2)<<16));


    if(channel==G35_DDC_NARROW)
    {
        bar0_write32(REG_DDC1_FIR_CTRL1+reg_offset,(params->fir_len>>1) | (((uint32_t)params->fir_control)<<16));
    }
    else
    {
        bar0_write32(REG_DDC1_FIR_CTRL1+reg_offset,(params->fir_len) | (((uint32_t)params->fir_control)<<16));
    }


    address=(channel<<28);

    bar0_write32(REG_FIR_COEFF,address | (1<<27));

    for(i=num_coeffs-1;i>=0;i--)
    {
        bar0_write32(REG_FIR_COEFF,address | (coeffs[i]&0x3FFFF));
    }
}


void g35_pci_ddc1_enable(struct g35ddc_pci *dev,int channel)
{
 struct g35_ddc1_params *params;
 uint32_t reg_offset;

    params=dev->g35_dev.ddc1_params+channel;

    reg_offset=channel*REG_DDC1_COUNT*4;

    bar0_write32(REG_DDC1_CTRL1+reg_offset,params->control_word);
    bar0_read32(REG_DDC1_CTRL1+reg_offset);
}

void g35_pci_ddc1_disable(struct g35ddc_pci *dev,int channel)
{
 struct g35_ddc1_params *params;
 uint32_t reg_offset;

    params=dev->g35_dev.ddc1_params+channel;

    reg_offset=channel*REG_DDC1_COUNT*4;

    bar0_write32(REG_DDC1_CTRL1+reg_offset,params->control_word & 0xFFFE);
}

static int g35_pci_ddc1_prepare(struct g35ddc_pci *dev,uint32_t channel,int32_t *coeffs,uint32_t num_coeffs)
{
 int status;

    status=g35_pci_enable_dma_channel(dev,G35_IO_STREAM_IN_DDC+channel);

    if(unlikely(status!=0))
    {
        return status;
    }


    if(channel==G35_DDC_WIDE)
    {
         if(dev->g35_dev.io_stream[G35_IO_STREAM_IN_DDC_WIDE].bytes_per_sec>100000000)
         {
               bar0_write32(REG_WDDC_SDRAM_BUFF,0x01|0x02);
         }
         else
         {
               bar0_write32(REG_WDDC_SDRAM_BUFF,0x01);
         }

         bar0_read32(REG_WDDC_SDRAM_BUFF);
    }

    g35_pci_ddc1_setup(dev,channel,coeffs,num_coeffs);

    return 0;
}

static int g35_pci_ddc1_start(struct g35ddc_pci *dev,uint32_t channel)
{
    if(!unlikely(dev->g35_dev.io_stream[G35_IO_STREAM_IN_DDC+channel].active))
    {
        return -EPERM;
    }

    g35_pci_ddc1_enable(dev,channel);

    return 0;
}

static void g35_pci_ddc1_stop(struct g35ddc_pci *dev,uint32_t channel)
{
 int stream_index;

    stream_index=G35_IO_STREAM_IN_DDC+channel;

    if(likely(dev->g35_dev.io_stream[stream_index].active))
    {
        g35_pci_disable_dma_channel(dev,stream_index);

        g35_pci_ddc1_disable(dev,channel);
    }
}

static void g35_pci_ddc1_update_tuning_word(struct g35ddc_pci *dev,uint32_t channel)
{
 struct g35_ddc1_params *params;
 uint32_t reg_offset;

    params=dev->g35_dev.ddc1_params+channel;

    reg_offset=channel*REG_DDC1_COUNT*4;

    bar0_write32(REG_DDC1_TUNING_WORD+reg_offset,params->tuning_word);
    bar0_read32(REG_DDC1_TUNING_WORD+reg_offset);
}

static int g35_pci_ddc1_update_frequency(struct g35ddc_pci *dev,uint32_t channel)
{
    g35_pci_ddc1_update_tuning_word(dev,channel);
    return 0;
}


static int g35_pci_ddc1_update_phase_shift(struct g35ddc_pci *dev,uint32_t channel)
{
 struct g35_ddc1_params *params;
 uint32_t reg_offset;

    params=dev->g35_dev.ddc1_params+channel;

    reg_offset=channel*REG_DDC1_COUNT*4;

    bar0_write32(REG_DDC1_START_PHASE+reg_offset,params->phase_shift);

    return 0;
}

static int g35_pci_wd(struct g35ddc_pci *dev)
{
    return ((bar4_read32(GPIO_INPUT_VALUE)>>1)&0x7F)>=68;
}

static int g35_pci_read_config(struct g35ddc_pci *dev)
{
 uint32_t u32;
 uint32_t ctl;
 int i;
 uint8_t to_read;
 uint32_t size;
 uint8_t *p;
 int status;

    status=-EIO;

    size=8192;
    p=dev->g35_dev.config+4;

    u32=bar4_read32(LB_CTL);

    if((u32 & 0x10000)==0)
    {
        u32|=0x10000;
        bar4_write32(LB_CTL,u32);
    }

    ctl = (G35_I2C_DIV_A << 14) | (G35_I2C_DIV_B << 8) | 0x4E;

    bar4_write32(TWI_CTRL,ctl);

    ctl=bar4_read32(TWI_CTRL);

    for(i=0;i<G35_I2C_MAX_RETRY;i++)
    {
        if((bar4_read32(TWI_STATUS) & 0x0100)==0)
        {
            break;
        }
    }

    if(unlikely(i>=G35_I2C_MAX_RETRY))
    {
        G35_ERROR("read eeprom:error:i2c is busy");
        goto error_exit;
    }

    udelay(100);

    bar4_read32(TWI_IRT_STATUS);

    bar4_write32(TWI_DATA,0);
    bar4_write32(TWI_DATA,0);

    bar4_write32(TWI_ADDRESS,0x51);

    for(i=0;i<G35_I2C_MAX_RETRY;i++)
    {
        u32=bar4_read32(TWI_IRT_STATUS);

        if(u32 & 0x01)
        {
            break;
        }

        if(u32 & 0x04)
        {
            G35_ERROR("read eeprom:i2c:addr:NACK");
            goto error_exit;
        }

        if(u32 & 0x08)
        {
            G35_ERROR("read eeprom:i2c:addr:transfer timeout");
            goto error_exit;
        }
    }

    if(unlikely(i>=G35_I2C_MAX_RETRY))
    {
        G35_ERROR("read eeprom:i2c:addr:max retry");
        goto error_exit;
    }

    ctl|=0x01;

    bar4_write32(TWI_CTRL,ctl);

	while(size!=0)
    {
        if(size>G35_I2C_FIFO_SIZE)
        {
            to_read=G35_I2C_FIFO_SIZE;
            size-=G35_I2C_FIFO_SIZE;
        }
        else
        {
            to_read=size;
            size=0;
        }

        bar4_write32(TWI_TR_SIZE,to_read);

        bar4_write32(TWI_ADDRESS,0x51);

	    for(i=0;i<G35_I2C_MAX_RETRY;i++)
	    {
	        if(bar4_read32(TWI_IRT_STATUS) & 0x01)
	        {
	            break;
	        }
	    }

	    if(unlikely(i>=G35_I2C_MAX_RETRY))
	    {
	        G35_ERROR("read eeprom:i2c:data:transfer completion not detected");
	        goto error_exit;
	    }

        while(to_read--)
        {
            *p++=(uint8_t)(bar4_read32(TWI_DATA) & 0xFF);
        }
    }

    status=0;

    error_exit:

    bar4_write32(TWI_CTRL,ctl | 0x01);
    bar4_read32(TWI_CTRL);
    bar4_write32(LB_CTL,bar4_read32(LB_CTL) & (~0x10000));
    bar4_read32(LB_CTL);

    return status;
}

static int g35_pci_ddc1_playback_start(struct g35ddc_pci *dev,uint32_t *fir_coeffs,uint32_t num_coeffs)
{
 int status;

    g35_pci_ddc1_setup(dev,G35_DDC_WIDE,fir_coeffs,num_coeffs);
    g35_pci_ddc1_enable(dev,G35_DDC_WIDE);

    status=g35_pci_enable_dma_channel(dev,G35_IO_STREAM_IN_DDC_PLAYBACK);

    if(likely(status==0))
    {
        status=g35_pci_enable_dma_channel(dev,G35_IO_STREAM_OUT_DDC_PLAYBACK);

        if(unlikely(status!=0))
        {
            g35_pci_disable_dma_channel(dev,G35_IO_STREAM_IN_DDC_PLAYBACK);
            g35_pci_ddc1_disable(dev,G35_DDC_WIDE);
        }
    }
    else
    {
        g35_pci_ddc1_disable(dev,G35_DDC_WIDE);
    }

    return status;
}

static void g35_pci_ddc1_playback_stop(struct g35ddc_pci *dev)
{
    if(dev->g35_dev.io_stream[G35_IO_STREAM_IN_DDC_PLAYBACK].active)
    {
        g35_pci_disable_dma_channel(dev,G35_IO_STREAM_IN_DDC_PLAYBACK);
        g35_pci_disable_dma_channel(dev,G35_IO_STREAM_OUT_DDC_PLAYBACK);

        g35_pci_ddc1_disable(dev,G35_DDC_WIDE);
    }
}


static int g35_pci_dac_start(struct g35ddc_pci *dev,uint32_t dac_control)
{
 int status;

    status=g35_pci_enable_dma_channel(dev,G35_IO_STREAM_OUT_DAC);

    if(unlikely(status))
    {
        return status;
    }

    bar0_write32(REG_DAC_CONTROL,dac_control);

    return 0;
}

static void g35_pci_dac_stop(struct g35ddc_pci *dev)
{
    if(dev->g35_dev.io_stream[G35_IO_STREAM_OUT_DAC].active)
    {
        g35_pci_disable_dma_channel(dev,G35_IO_STREAM_OUT_DAC);

        if(dev->g35_dev.state.fpga)
        {
            bar0_write32(REG_DAC_CONTROL,0);
        }
    }
}

static void g35_pci_cancel_buffers(struct g35ddc_pci *dev,int stream_index)
{
 struct g35_io_stream *io_stream;
 struct g35_io_buffer *io_buffer;
 struct g35_dma_channel *channel;

    io_stream=&dev->g35_dev.io_stream[stream_index];
    channel=&dev->dma_channel[stream_index];

    g35_mutex_lock(&channel->mutex);

    while((io_buffer=g35_io_stream_get_pending_buffer_nolock(io_stream)))
    {
        g35_io_buffer_sg_unmap(io_buffer);
        g35_io_stream_put_complete_buffer(io_stream,io_buffer);
    }

    g35_mutex_unlock(&channel->mutex);
}


static int g35_pci_coh_arm(struct g35ddc_pci *dev,uint8_t master)
{
    bar0_write32(REG_COHERENT_MODE_CTRL,(master?0x01:0x00) | 0x02);
    bar0_read32(REG_COHERENT_MODE_CTRL);

    return 0;
}

static int g35_pci_coh_trigger(struct g35ddc_pci *dev)
{
    bar0_write32(REG_COHERENT_MODE_CTRL,0x01 | 0x04 | 0x02);
    bar0_read32(REG_COHERENT_MODE_CTRL);

    return 0;
}

static int g35_pci_coh_set(struct g35ddc_pci *dev,uint8_t value)
{
    if(dev->g35_dev.config[2]<0x35 || dev->g35_dev.state.power)
    {
        return -EPERM;
    }

    if(value)
    {
        gpio_set(GPIO_BIT_COH_OUT);
    }
    else
    {
        gpio_clr(GPIO_BIT_COH_OUT);
    }

    return 0;
}

static int g35_pci_coh_get(struct g35ddc_pci *dev,uint8_t *value)
{
    if(dev->g35_dev.config[2]<0x35 || dev->g35_dev.state.power)
    {
        return -EPERM;
    }

    *value=(bar4_read32(GPIO_INPUT_VALUE)&GPIO_BIT_COH_IN)?1:0;

    return 0;
}

static int g35_pci_ddc1_get_timestamp(struct g35ddc_pci *dev,uint32_t channel,union g35_timestamp *timestamp)
{
 uint32_t reg;

    reg=REG_DDC1_TIMESTAMP_LO+8*channel;

    timestamp->lo=bar0_read32(reg);
    timestamp->hi=bar0_read32(reg+4);

    return 0;
}

static int g35_pci_initialize(struct g35ddc_pci *dev)
{
 uint32_t i;
 uint8_t j,k;
 int status;

    g35_pci_setup(dev);

    bar4_write32(GPIO_DIRECTION_MODE,GPIO_BIT_POWER_STATE|GPIO_BIT_DMA_IRQ|GPIO_BIT_TMP|
            GPIO_BIT_TMP_ERROR|GPIO_BIT_FAN_ERROR);

    gpio_clr(GPIO_BIT_FE_POWER|GPIO_BIT_CTL|GPIO_BIT_FPGA_RESET);

    bar4_write32(GPIO_OUTPUT_ENABLE,GPIO_BIT_FE_POWER|GPIO_BIT_CTL|GPIO_BIT_FPGA_RESET);

    gpio_clr(GPIO_BIT_FE_POWER|GPIO_BIT_CTL|GPIO_BIT_FPGA_RESET);

    msleep(5);

    gpio_set(GPIO_BIT_CTL);

    msleep(15);

    i=(bar4_read32(GPIO_INPUT_VALUE)>>1) & 0x7F;

    if(i==0x64)
    {
        dev->g35_dev.config[1]=1;

        bar4_write32(GPIO_DIRECTION_MODE,GPIO_BIT_POWER_STATE|GPIO_BIT_DMA_IRQ|GPIO_BIT_TMP);

        gpio_clr(GPIO_BIT_FAN_ERROR|GPIO_BIT_TMP_ERROR);

        bar4_write32(GPIO_OUTPUT_ENABLE,GPIO_BIT_FE_POWER|GPIO_BIT_CTL|GPIO_BIT_FPGA_RESET|GPIO_BIT_FAN_ERROR|GPIO_BIT_TMP_ERROR);

        j=0x5B;
        i=bar4_read32(GPIO_OUTPUT_VALUE);

        for(k=0;k<8;k++,j<<=1)
        {
            if(j&0x80)
            {
                i|=GPIO_BIT_FAN_ERROR;
            }
            else
            {
                i&=~GPIO_BIT_FAN_ERROR;
            }

            bar4_write32(GPIO_OUTPUT_VALUE,i);
            i=bar4_read32(GPIO_OUTPUT_VALUE);

            msleep(2);

            i|=GPIO_BIT_TMP_ERROR;
            bar4_write32(GPIO_OUTPUT_VALUE,i);
            i=bar4_read32(GPIO_OUTPUT_VALUE);

            msleep(3);

            i&=~GPIO_BIT_TMP_ERROR;
            bar4_write32(GPIO_OUTPUT_VALUE,i);
            i=bar4_read32(GPIO_OUTPUT_VALUE);

            msleep(2);
        }

        msleep(4);

        dev->g35_dev.config[2]=(bar4_read32(GPIO_INPUT_VALUE)>>1)&0x7F;

        i&=~(GPIO_BIT_TMP_ERROR|GPIO_BIT_FAN_ERROR);
        bar4_write32(GPIO_OUTPUT_VALUE,i);
        bar4_read32(GPIO_OUTPUT_VALUE);

        msleep(3);

        i&=~GPIO_BIT_CTL;
        bar4_write32(GPIO_OUTPUT_VALUE,i);
        bar4_read32(GPIO_OUTPUT_VALUE);

        msleep(1);

        if(dev->g35_dev.config[2]>=0x35)
        {
            bar4_write32(GPIO_OUTPUT_ENABLE,GPIO_BIT_FE_POWER|GPIO_BIT_FPGA_RESET|GPIO_BIT_CTL|GPIO_BIT_COH_OUT);

            bar4_write32(GPIO_DIRECTION_MODE,GPIO_BIT_POWER_STATE|GPIO_BIT_DMA_IRQ|GPIO_BIT_TMP|
                    GPIO_BIT_TMP_ERROR);
        }
        else
        {
            bar4_write32(GPIO_OUTPUT_ENABLE,GPIO_BIT_FE_POWER|GPIO_BIT_FPGA_RESET|GPIO_BIT_CTL);

            bar4_write32(GPIO_DIRECTION_MODE,GPIO_BIT_POWER_STATE|GPIO_BIT_DMA_IRQ|GPIO_BIT_TMP|
                    GPIO_BIT_TMP_ERROR|GPIO_BIT_FAN_ERROR);
        }
    }
    else
    {
        bar4_write32(GPIO_OUTPUT_ENABLE,GPIO_BIT_FE_POWER|GPIO_BIT_FPGA_RESET);

        bar4_write32(GPIO_DIRECTION_MODE,GPIO_BIT_POWER_STATE|GPIO_BIT_DMA_IRQ|GPIO_BIT_TMP|
            GPIO_BIT_TMP_ERROR|GPIO_BIT_FAN_ERROR|GPIO_BIT_CTL);
    }


    for(i=0;i<3;i++)
    {
        status=g35_pci_read_config(dev);

        if(status==0)
        {
            break;
        }
        mdelay(100);
    }

    if(unlikely(status))
    {
        G35_ERROR("failed to read configuration data!\n");
        return status;
    }

    dev->g35_dev.func_delete=(g35_func_delete_t)g35_pci_delete;
    dev->g35_dev.func_open=(g35_func_open_t)g35_pci_open;
    dev->g35_dev.func_close=(g35_func_close_t)g35_pci_close;
    dev->g35_dev.func_set_power=(g35_func_set_power_t)g35_pci_set_power;
    dev->g35_dev.func_configure_fpga=(g35_func_configure_fpga_t)g35_pci_configure_fpga;
    dev->g35_dev.func_rf_control1=(g35_func_rf_control)g35_pci_rf_control1;
    dev->g35_dev.func_rf_control2=(g35_func_rf_control)g35_pci_rf_control2;
    dev->g35_dev.func_set_noise_blanker=(g35_func_set_noise_blanker_t)g35_pci_set_noise_blanker;
    dev->g35_dev.func_put_buffer=(g35_func_put_buffer_t)g35_pci_put_buffer;
    dev->g35_dev.func_if_start=(g35_func_if_start_t)g35_pci_if_start;
    dev->g35_dev.func_if_stop=(g35_func_if_stop_t)g35_pci_if_stop;
    dev->g35_dev.func_get_adc_level=(g35_func_get_adc_level_t)g35_pci_get_adc_level;
    dev->g35_dev.func_mmap=(g35_func_mmap_t)g35_pci_mmap;
    dev->g35_dev.func_wd=(g35_func_wd_t)g35_pci_wd;
    dev->g35_dev.func_ddc1_prepare=(g35_func_ddc1_prepare_t)g35_pci_ddc1_prepare;
    dev->g35_dev.func_ddc1_stop=(g35_func_ddc1_stop_t)g35_pci_ddc1_stop;
    dev->g35_dev.func_ddc1_start=(g35_func_ddc1_start_t)g35_pci_ddc1_start;
    dev->g35_dev.func_ddc1_update_frequency=(g35_func_ddc1_update_frequency_t)g35_pci_ddc1_update_frequency;
    dev->g35_dev.func_ddc1_update_phase_shift=(g35_func_ddc1_update_phase_shift_t)g35_pci_ddc1_update_phase_shift;
    dev->g35_dev.func_ddc1_playback_start=(g35_func_ddc1_playback_start_t)g35_pci_ddc1_playback_start;
    dev->g35_dev.func_ddc1_playback_stop=(g35_func_ddc1_playback_stop_t)g35_pci_ddc1_playback_stop;
    dev->g35_dev.func_dac_start=(g35_func_dac_start_t)g35_pci_dac_start;
    dev->g35_dev.func_dac_stop=(g35_func_dac_stop_t)g35_pci_dac_stop;
    dev->g35_dev.func_cancel_buffers=(g35_func_cancel_buffers_t)g35_pci_cancel_buffers;
    dev->g35_dev.func_coh_arm=(g35_func_coh_arm_t)g35_pci_coh_arm;
    dev->g35_dev.func_coh_trigger=(g35_func_coh_trigger_t)g35_pci_coh_trigger;
    dev->g35_dev.func_coh_set=(g35_func_coh_set_t)g35_pci_coh_set;
    dev->g35_dev.func_coh_get=(g35_func_coh_get_t)g35_pci_coh_get;
    dev->g35_dev.func_ddc1_get_timestamp=(g35_func_ddc1_get_timestamp_t)g35_pci_ddc1_get_timestamp;

    spin_lock_init(&dev->irq_spinlock);

    INIT_WORK(&dev->irq_work,g35_pci_irq_work_handler);

    for(i=0;i<G35_IO_STREAM_COUNT;i++)
    {
        if(i>=G35_IO_STREAM_OUT_FIRST && i<=G35_IO_STREAM_OUT_LAST)
        {
            dev->dma_channel[i].index=i-G35_IO_STREAM_OUT_FIRST;
            dev->dma_channel[i].out=1;
        }
        else
        {
            dev->dma_channel[i].index=i-G35_IO_STREAM_IN_FIRST;
            dev->dma_channel[i].out=0;
        }

        g35_mutex_init(&dev->dma_channel[i].mutex);
    }

    dev->dma_sg_list_memory_size=((G35_DMA_PAGE_SIZE*(dev->dma_64?7:6)+PAGE_SIZE-1)/PAGE_SIZE)*PAGE_SIZE;
    dev->dma_max_buffer_size=(dev->dma_sg_list_memory_size/
            (dev->dma_64?sizeof(struct g35_sg_segment_64):sizeof(struct g35_sg_segment_32))-1)*PAGE_SIZE;


    return 0;
}

#ifdef CONFIG_PM

static int g35_pci_suspend(struct pci_dev *pci_dev,pm_message_t state)
{
 struct g35ddc_pci *dev;

    dev=pci_get_drvdata(pci_dev);

    if(likely(dev!=NULL))
    {
        G35_DEV_LOCK2;

        if(dev->g35_dev.state.power)
        {
            dev->g35_dev.state.suspended=1;
            g35_set_power(&dev->g35_dev,0,0);
        }

        pci_save_state(pci_dev);

        if(pci_set_power_state(pci_dev,pci_choose_state(pci_dev,state))!=0)
        {
            pci_disable_device(pci_dev);
            dev->disabled=1;
        }

        G35_DEV_UNLOCK2;
    }

    return 0;
}

static int g35_pci_resume(struct pci_dev *pci_dev)
{
 struct g35ddc_pci *dev;
 int state;

    dev=pci_get_drvdata(pci_dev);

    if(likely(dev!=NULL))
    {
        if(dev->disabled)
        {
            state=pci_enable_device(pci_dev);

            if(unlikely(state!=0))
            {
                return state;
            }

            dev->disabled=0;
        }

        state=pci_set_power_state(pci_dev,PCI_D0);

        if(unlikely(state!=0))
        {
            pci_disable_device(pci_dev);
            dev->disabled=1;
            return state;
        }

        pci_restore_state(pci_dev);

        g35_pci_setup(dev);

        if(likely(dev->g35_dev.config[1]))
        {
            if(dev->g35_dev.config[2]>=0x35)
            {
                bar4_write32(GPIO_OUTPUT_ENABLE,GPIO_BIT_FE_POWER|GPIO_BIT_FPGA_RESET|GPIO_BIT_CTL|GPIO_BIT_COH_OUT);

                bar4_write32(GPIO_DIRECTION_MODE,GPIO_BIT_POWER_STATE|GPIO_BIT_DMA_IRQ|GPIO_BIT_TMP|
                        GPIO_BIT_TMP_ERROR);
            }
            else
            {
                bar4_write32(GPIO_OUTPUT_ENABLE,GPIO_BIT_FE_POWER|GPIO_BIT_FPGA_RESET|GPIO_BIT_CTL);

                bar4_write32(GPIO_DIRECTION_MODE,GPIO_BIT_POWER_STATE|GPIO_BIT_DMA_IRQ|GPIO_BIT_TMP|
                        GPIO_BIT_TMP_ERROR|GPIO_BIT_FAN_ERROR);
            }
        }
        else
        {
            bar4_write32(GPIO_OUTPUT_ENABLE,GPIO_BIT_FE_POWER|GPIO_BIT_FPGA_RESET);

            bar4_write32(GPIO_DIRECTION_MODE,GPIO_BIT_POWER_STATE|GPIO_BIT_DMA_IRQ|GPIO_BIT_TMP|
                GPIO_BIT_TMP_ERROR|GPIO_BIT_FAN_ERROR|GPIO_BIT_CTL);
        }
    }

    return 0;
}

#endif //CONFIG_PM


static struct pci_device_id g35_pci_id_table[]=
{
    {.vendor=G35_PCI_VENDOR,.device=G35_PCI_DEVICE,.subvendor=G35_PCI_SUBSYSTEM_VENDOR,.subdevice=G35_PCI_SUBSYSTEM_DEVICE},
    {},
};

static struct pci_driver g35_pci_driver=
{
    .name=G35_DEVICE_NAME,
    .probe=g35_pci_probe,
    .remove=g35_pci_remove,
    .id_table=g35_pci_id_table,
#ifdef CONFIG_PM
    .suspend=g35_pci_suspend,
    .resume=g35_pci_resume
#endif
};

MODULE_DEVICE_TABLE(pci,g35_pci_id_table);

#endif //__G35_PCI__

static int g35_fo_open(struct inode *node, struct file *file)
{
 uint32_t index;
 int status;
 struct g35ddc *dev;

    index=iminor(file->f_path.dentry->d_inode);

    G35_REG_LOCK;

    dev=g35_registered_devices[index];

    if(unlikely(dev==NULL))
    {
        G35_REG_UNLOCK;
        return -ENXIO;
    }

    get_device(&dev->dev);

    G35_DEV_LOCK;

    if(unlikely(dev->state.open))
    {
        G35_DEV_UNLOCK;
        G35_REG_UNLOCK;
        put_device(&dev->dev);
        return -EBUSY;
    }

    dev->state.open=1;

    G35_DEV_UNLOCK;

    dev->workqueue=create_singlethread_workqueue(G35_DEVICE_NAME);

    if(unlikely(dev->workqueue==NULL))
    {
        dev->state.open=0;
        G35_REG_UNLOCK;
        put_device(&dev->dev);
        return -EIO;
    }

    G35_DEV_LOCK;

    status=dev->func_open(dev);

    if(unlikely(status<0))
    {
        dev->state.open=0;
    }

    G35_DEV_UNLOCK;

    if(likely(status>=0))
    {
        file->private_data=dev;

        dev->wd_counter=0;

        dev->wd_timer.expires=jiffies+2*HZ;
        add_timer(&dev->wd_timer);
    }
    else
    {
        destroy_workqueue(dev->workqueue);
        dev->workqueue=NULL;
    }

    G35_REG_UNLOCK;

    if(unlikely(status<0))
    {
        put_device(&dev->dev);
    }

    return status;
}

static int g35_fo_release(struct inode *node, struct file *file)
{
 struct g35ddc *dev;

    dev=file->private_data;

    if(likely(dev!=NULL))
    {
        G35_DEV_LOCK;

        g35_set_power(dev,0,0);

        dev->state.open=0;

        dev->func_close(dev);

        wake_up_all(&dev->power_wait_queue);

        G35_DEV_UNLOCK;

        del_timer_sync(&dev->wd_timer);

        if(likely(dev->workqueue!=NULL))
        {
            flush_workqueue(dev->workqueue);
            destroy_workqueue(dev->workqueue);
            dev->workqueue=NULL;
        }

        put_device(&dev->dev);

        file->private_data=NULL;
    }

    return 0;
}



static int __init g35_init(void)
{
 int status;

    g35_mutex_init(&g35_reg_mutex);

    printk(KERN_INFO G35_DESCRIPTION "\n");

    status=alloc_chrdev_region(&g35_dev0,0,G35_MAX_DEVICE_COUNT,G35_DEVICE_NAME);

    if(status<0)
    {
        G35_ERROR("unable to allocate chrdev region!\n");
        goto error_mutex_destroy;
    }

    status=class_register(&g35_class);

    if(status<0)
    {
        G35_ERROR("failed to register class!\n");
        goto error_unregister_chrdev_region;
    }


#ifdef __G35_PCI__
    status=pci_register_driver(&g35_pci_driver);

    if(unlikely(status<0))
    {
        G35_ERROR("failed to register pci driver!\n");
        goto error_usb_deregister;
    }
#endif //__G35_PCI__

    return 0;

    error_usb_deregister:
        class_unregister(&g35_class);

    error_unregister_chrdev_region:
        unregister_chrdev_region(g35_dev0,G35_MAX_DEVICE_COUNT);

    error_mutex_destroy:
        g35_mutex_destroy(&g35_reg_mutex);

    return status;
}

static void __exit g35_exit(void)
{
#ifdef __G35_PCI__
    pci_unregister_driver(&g35_pci_driver);
#endif //__G35_PCI__

    class_unregister(&g35_class);
    unregister_chrdev_region(g35_dev0,G35_MAX_DEVICE_COUNT);

    g35_mutex_destroy(&g35_reg_mutex);
}


static int g35_fo_mmap(struct file *file,struct vm_area_struct *vma)
{
 struct g35ddc *dev;

    dev=file->private_data;

    if(unlikely(dev==NULL || !dev->state.connected))
    {
        return -ENODEV;
    }

    return dev->func_mmap(dev,vma);
}

#if defined(HAVE_UNLOCKED_IOCTL) || LINUX_VERSION_CODE>=KERNEL_VERSION(5,0,0)
static long g35_fo_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
#else
static int g35_fo_ioctl(struct inode *node,struct file *file, unsigned int cmd, unsigned long arg)
#endif //HAVE_UNLOCKED_IOCTL
{
 int status;
 struct g35ddc *dev;
 uint8_t byte;
 uint16_t word;
 uint32_t dword;
 union g35_rf_control rf;
 struct g35_fpga_data fpga;
 struct g35_ddc1_start_params ddc1_start_params;
 struct g35_dac_params dac_params;

    dev=file->private_data;

    if(unlikely(dev==NULL))
    {
        return -ENODEV;
    }

    switch(cmd)
    {
        case G35_IOCTL_DDC1_0_PUT_BUFFER:
            status=g35_put_buffer(dev,G35_IO_STREAM_IN_DDC,(void __user*)arg,1);
            break;

        case G35_IOCTL_DDC1_1_PUT_BUFFER:
            status=g35_put_buffer(dev,G35_IO_STREAM_IN_DDC_WIDE,(void __user*)arg,1);
            break;

        case G35_IOCTL_IF_PUT_BUFFER:
            status=g35_put_buffer(dev,G35_IO_STREAM_IN_IF,(void __user*)arg,1);
            break;

        case G35_IOCTL_DAC_PUT_BUFFER:
            status=g35_put_buffer(dev,G35_IO_STREAM_OUT_DAC,(void __user*)arg,0);
            break;

        case G35_IOCTL_DDC1_0_GET_BUFFER:
            status=g35_get_complete_buffer(dev,G35_IO_STREAM_IN_DDC,arg);
            break;

        case G35_IOCTL_DDC1_1_GET_BUFFER:
            status=g35_get_complete_buffer(dev,G35_IO_STREAM_IN_DDC_WIDE,arg);
            break;

        case G35_IOCTL_IF_GET_BUFFER:
            status=g35_get_complete_buffer(dev,G35_IO_STREAM_IN_IF,arg);
            break;

        case G35_IOCTL_DAC_GET_BUFFER:
            status=g35_get_complete_buffer(dev,G35_IO_STREAM_OUT_DAC,arg);
            break;

        case G35_IOCTL_DDC_PLAYBACK_PUT_BUFFER_OUT:
            status=g35_put_buffer(dev,G35_IO_STREAM_OUT_DDC_PLAYBACK,(void __user*)arg,0);
            break;

        case G35_IOCTL_DDC_PLAYBACK_GET_BUFFER_OUT:
            status=g35_get_complete_buffer(dev,G35_IO_STREAM_OUT_DDC_PLAYBACK,arg);
            break;

        case G35_IOCTL_DDC_PLAYBACK_PUT_BUFFER_IN:
            status=g35_put_buffer(dev,G35_IO_STREAM_IN_DDC_PLAYBACK,(void __user*)arg,1);
            break;

        case G35_IOCTL_DDC_PLAYBACK_GET_BUFFER_IN:
            status=g35_get_complete_buffer(dev,G35_IO_STREAM_IN_DDC_PLAYBACK,arg);
            break;

        case G35_IOCTL_GET_ADC_LEVEL:
            status=g35_get_adc_level(dev,&word);

            if(likely(status==0))
            {
                status=put_user(word,(uint16_t __user*)arg);
            }
            break;

        case G35_IOCTL_DDC_PLAYBACK_SET_FREQUENCY:
            if(unlikely(get_user(dword,(uint32_t __user*)compat_ptr(arg))<0))
            {
                return -EFAULT;
            }

            status=g35_ddc_playback_set_frequency(dev,dword);
            break;

        case G35_IOCTL_DDC1_0_SET_FREQUENCY:
            if(unlikely(get_user(dword,(uint32_t __user*)arg)<0))
            {
                return -EFAULT;
            }

            status=g35_ddc1_set_frequency(dev,0,dword);
            break;

        case G35_IOCTL_DDC1_0_SET_PHASE_SHIFT:
            if(unlikely(get_user(dword,(uint32_t __user*)arg)<0))
            {
                return -EFAULT;
            }

            status=g35_ddc1_set_phase_shift(dev,0,dword);
            break;

        case G35_IOCTL_DDC1_1_SET_FREQUENCY:
            if(unlikely(get_user(dword,(uint32_t __user*)arg)<0))
            {
                return -EFAULT;
            }

            status=g35_ddc1_set_frequency(dev,1,dword);
            break;

        case G35_IOCTL_DDC1_1_SET_PHASE_SHIFT:
            if(unlikely(get_user(dword,(uint32_t __user*)arg)<0))
            {
                return -EFAULT;
            }

            status=g35_ddc1_set_phase_shift(dev,1,dword);
            break;

        case G35_IOCTL_DDC1_0_PREPARE:
            if(unlikely(copy_from_user(&ddc1_start_params,(void __user*)arg,sizeof(ddc1_start_params))!=0))
            {
                return -EFAULT;
            }

            status=g35_ddc1_prepare(dev,0,&ddc1_start_params.params,(void __user*)ddc1_start_params.coeffs,ddc1_start_params.num_coeffs,
                    ddc1_start_params.buffer_size,ddc1_start_params.bytes_per_sec);
            break;

        case G35_IOCTL_DDC1_0_START:
            status=g35_ddc1_start(dev,0);
            break;

        case G35_IOCTL_DDC1_0_STOP:
            g35_ddc1_stop(dev,0);
            status=0;
            break;

        case G35_IOCTL_DDC1_1_PREPARE:
            if(unlikely(copy_from_user(&ddc1_start_params,(void __user*)arg,sizeof(ddc1_start_params))!=0))
            {
                return -EFAULT;
            }

            status=g35_ddc1_prepare(dev,1,&ddc1_start_params.params,(void __user*)ddc1_start_params.coeffs,ddc1_start_params.num_coeffs,ddc1_start_params.buffer_size,ddc1_start_params.bytes_per_sec);
            break;

        case G35_IOCTL_DDC1_1_START:
            status=g35_ddc1_start(dev,1);
            break;

        case G35_IOCTL_DDC1_1_STOP:
            g35_ddc1_stop(dev,1);
            status=0;
            break;

        case G35_IOCTL_DAC_START:
            if(unlikely(copy_from_user(&dac_params,(void __user*)arg,sizeof(dac_params))!=0))
            {
                return -EFAULT;
            }

            status=g35_dac_start(dev,&dac_params);
            break;

        case G35_IOCTL_DAC_STOP:
            g35_dac_stop(dev);
            status=0;
            break;

        case G35_IOCTL_DDC_PLAYBACK_START:
            if(unlikely(copy_from_user(&ddc1_start_params,(void __user*)arg,sizeof(ddc1_start_params))!=0))
            {
                return -EFAULT;
            }

            status=g35_ddc1_playback_start(dev,&ddc1_start_params.params,(void __user*)ddc1_start_params.coeffs,ddc1_start_params.buffer_size,ddc1_start_params.bytes_per_sec);
            break;

        case G35_IOCTL_DDC_PLAYBACK_STOP:
            g35_ddc1_playback_stop(dev);
            status=0;
            break;

        case G35_IOCTL_IF_START:
            if(unlikely(get_user(word,(uint16_t __user*)arg)<0))
            {
                return -EFAULT;
            }

            status=g35_if_start(dev,word);
            break;

        case G35_IOCTL_IF_STOP:
            g35_if_stop(dev);

            status=0;
            break;

        case G35_IOCTL_GET_STATE:
            status=g35_get_device_state(dev,(void __user*)arg);
            break;

        case G35_IOCTL_SET_NOISE_BLANKER:
            if(unlikely(get_user(word,(uint16_t __user*)arg)<0))
            {
                return -EFAULT;
            }

            status=g35_set_noise_blanker(dev,word);
            break;

        case G35_IOCTL_IF_CANCEL_BUFFERS:
            g35_cancel_buffers(dev,G35_IO_STREAM_IN_IF);
            status=0;
            break;

        case G35_IOCTL_DDC1_0_CANCEL_BUFFERS:
            g35_cancel_buffers(dev,G35_IO_STREAM_IN_DDC);
            status=0;
            break;

        case G35_IOCTL_DDC1_1_CANCEL_BUFFERS:
            g35_cancel_buffers(dev,G35_IO_STREAM_IN_DDC_WIDE);
            status=0;
            break;

        case G35_IOCTL_CANCEL_BUFFERS_PLAYBACK_IN:
            g35_cancel_buffers(dev,G35_IO_STREAM_IN_DDC_PLAYBACK);
            status=0;
            break;

        case G35_IOCTL_CANCEL_BUFFERS_PLAYBACK_OUT:
            g35_cancel_buffers(dev,G35_IO_STREAM_OUT_DDC_PLAYBACK);
            status=0;
            break;

        case G35_IOCTL_SET_POWER:
            if(unlikely(get_user(byte,(uint8_t __user*)arg)<0))
            {
                return -EFAULT;
            }

            status=g35_set_power(dev,byte,1);
            break;

        case G35_IOCTL_CONFIGURE_FPGA:
            if(unlikely(copy_from_user(&fpga,(struct g35_fpga_data __user*)arg,sizeof(fpga))!=0))
            {
                return -EFAULT;
            }

            status=g35_configure_fpga(dev,(void __user*)fpga.buffer,fpga.size);
            break;

        case G35_IOCTL_GET_CONFIG:
            status=copy_to_user((void __user*)arg,dev->config,sizeof(dev->config))?-EFAULT:0;
            break;

        case G35_IOCTL_WAIT_FOR_POWER_OFF:
            status=g35_wait_for_power_off(dev);
            break;

        case G35_IOCTL_RF_CONTROL1:
            if(unlikely(copy_from_user(&rf,(union g35_rf_control __user*)arg,sizeof(rf))!=0))
            {
                return -EFAULT;
            }

            status=g35_rf_control1(dev,rf.value);
            break;
        case G35_IOCTL_RF_CONTROL2:
            if(unlikely(copy_from_user(&rf,(union g35_rf_control __user*)arg,sizeof(rf))!=0))
            {
                return -EFAULT;
            }

            status=g35_rf_control2(dev,rf.value);
            break;

        case G35_IOCTL_DDC1_0_GET_TIMESTAMP:
            status=g35_ddc1_get_timestamp(dev,G35_DDC_NARROW,(void __user*)arg);
            break;

        case G35_IOCTL_DDC1_1_GET_TIMESTAMP:
            status=g35_ddc1_get_timestamp(dev,G35_DDC_WIDE,(void __user*)arg);
            break;

        case G35_IOCTL_COH_ARM:
            status=g35_coh_arm(dev,(void __user*)arg);
            break;

        case G35_IOCTL_COH_TRIGGER:
            status=g35_coh_trigger(dev);
            break;

        case G35_IOCTL_COH_SET:
            status=g35_coh_set(dev,(void __user*)arg);
            break;

        case G35_IOCTL_COH_GET:
            status=g35_coh_get(dev,(void __user*)arg);
            break;

        default:
            status=-EINVAL;
            break;
    }

    return status;
}

#if defined(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL)

static long g35_fo_ioctl_compat(struct file *file, unsigned int cmd, unsigned long arg)
{
 int status;
 struct g35ddc *dev;
 uint8_t byte;
 uint16_t word;
 uint32_t dword;
 union g35_rf_control rf;
 struct g35_fpga_data_compat fpga;
 struct g35_ddc1_start_params_compat ddc1_start_params;
 struct g35_dac_params dac_params;

    dev=file->private_data;

    if(unlikely(dev==NULL))
    {
        return -ENODEV;
    }

    switch(cmd)
    {
        case G35_IOCTL_DDC1_0_PUT_BUFFER_COMPAT:
            status=g35_put_buffer(dev,G35_IO_STREAM_IN_DDC,compat_ptr(arg),1);
            break;

        case G35_IOCTL_DDC1_1_PUT_BUFFER_COMPAT:
            status=g35_put_buffer(dev,G35_IO_STREAM_IN_DDC_WIDE,compat_ptr(arg),1);
            break;

        case G35_IOCTL_IF_PUT_BUFFER_COMPAT:
            status=g35_put_buffer(dev,G35_IO_STREAM_IN_IF,compat_ptr(arg),1);
            break;

        case G35_IOCTL_DAC_PUT_BUFFER_COMPAT:
            status=g35_put_buffer(dev,G35_IO_STREAM_OUT_DAC,compat_ptr(arg),0);
            break;

        case G35_IOCTL_DDC1_0_GET_BUFFER_COMPAT:
            status=g35_get_complete_buffer_compat(dev,G35_IO_STREAM_IN_DDC,arg);
            break;

        case G35_IOCTL_DDC1_1_GET_BUFFER_COMPAT:
            status=g35_get_complete_buffer_compat(dev,G35_IO_STREAM_IN_DDC_WIDE,arg);
            break;

        case G35_IOCTL_IF_GET_BUFFER_COMPAT:
            status=g35_get_complete_buffer_compat(dev,G35_IO_STREAM_IN_IF,arg);
            break;

        case G35_IOCTL_DAC_GET_BUFFER_COMPAT:
            status=g35_get_complete_buffer_compat(dev,G35_IO_STREAM_OUT_DAC,arg);
            break;

        case G35_IOCTL_DDC_PLAYBACK_PUT_BUFFER_OUT_COMPAT:
            status=g35_put_buffer(dev,G35_IO_STREAM_OUT_DDC_PLAYBACK,compat_ptr(arg),0);
            break;

        case G35_IOCTL_DDC_PLAYBACK_GET_BUFFER_OUT_COMPAT:
            status=g35_get_complete_buffer_compat(dev,G35_IO_STREAM_OUT_DDC_PLAYBACK,arg);
            break;

        case G35_IOCTL_DDC_PLAYBACK_PUT_BUFFER_IN_COMPAT:
            status=g35_put_buffer(dev,G35_IO_STREAM_OUT_DDC_PLAYBACK,compat_ptr(arg),1);
            break;

        case G35_IOCTL_DDC_PLAYBACK_GET_BUFFER_IN_COMPAT:
            status=g35_get_complete_buffer_compat(dev,G35_IO_STREAM_IN_DDC_PLAYBACK,arg);
            break;

        case G35_IOCTL_GET_ADC_LEVEL:
            status=g35_get_adc_level(dev,&word);

            if(likely(status==0))
            {
                status=put_user(word,(uint16_t __user*)compat_ptr(arg));
            }
            break;

        case G35_IOCTL_DDC_PLAYBACK_SET_FREQUENCY:
            if(unlikely(get_user(dword,(uint32_t __user*)compat_ptr(arg))<0))
            {
                return -EFAULT;
            }

            status=g35_ddc_playback_set_frequency(dev,dword);
            break;

        case G35_IOCTL_DDC1_0_SET_FREQUENCY:
            if(unlikely(get_user(dword,(uint32_t __user*)compat_ptr(arg))<0))
            {
                return -EFAULT;
            }

            status=g35_ddc1_set_frequency(dev,0,dword);
            break;

        case G35_IOCTL_DDC1_1_SET_FREQUENCY:
            if(unlikely(get_user(dword,(uint32_t __user*)compat_ptr(arg))<0))
            {
                return -EFAULT;
            }

            status=g35_ddc1_set_frequency(dev,1,dword);
            break;

        case G35_IOCTL_DDC1_0_PREPARE_COMPAT:
            if(unlikely(copy_from_user(&ddc1_start_params,(void __user*)compat_ptr(arg),sizeof(ddc1_start_params))!=0))
            {
                return -EFAULT;
            }

            status=g35_ddc1_prepare(dev,0,&ddc1_start_params.params,(void __user*)compat_ptr(ddc1_start_params.coeffs),ddc1_start_params.num_coeffs,ddc1_start_params.buffer_size,ddc1_start_params.bytes_per_sec);
            break;

        case G35_IOCTL_DDC1_0_START:
            status=g35_ddc1_start(dev,0);
            break;

        case G35_IOCTL_DDC1_0_STOP:
            g35_ddc1_stop(dev,0);
            status=0;
            break;

        case G35_IOCTL_DDC1_1_PREPARE_COMPAT:
            if(unlikely(copy_from_user(&ddc1_start_params,(void __user*)compat_ptr(arg),sizeof(ddc1_start_params))!=0))
            {
                return -EFAULT;
            }

            status=g35_ddc1_prepare(dev,1,&ddc1_start_params.params,(void __user*)compat_ptr(ddc1_start_params.coeffs),ddc1_start_params.num_coeffs,ddc1_start_params.buffer_size,ddc1_start_params.bytes_per_sec);
            break;

        case G35_IOCTL_DDC1_1_START:
            status=g35_ddc1_start(dev,1);
            break;

        case G35_IOCTL_DDC_PLAYBACK_START_COMPAT:
            if(unlikely(copy_from_user(&ddc1_start_params,(void __user*)compat_ptr(arg),sizeof(ddc1_start_params))!=0))
            {
                return -EFAULT;
            }

            status=g35_ddc1_playback_start(dev,&ddc1_start_params.params,(void __user*)compat_ptr(ddc1_start_params.coeffs),ddc1_start_params.buffer_size,ddc1_start_params.bytes_per_sec);
            break;

        case G35_IOCTL_DDC_PLAYBACK_STOP:
            g35_ddc1_playback_stop(dev);
            status=0;
            break;

        case G35_IOCTL_DDC1_1_STOP:
            g35_ddc1_stop(dev,1);
            status=0;
            break;

        case G35_IOCTL_DAC_START:
            if(unlikely(copy_from_user(&dac_params,(void __user*)compat_ptr(arg),sizeof(dac_params))!=0))
            {
                return -EFAULT;
            }

            status=g35_dac_start(dev,&dac_params);
            break;

        case G35_IOCTL_DAC_STOP:
            g35_dac_stop(dev);
            status=0;
            break;

        case G35_IOCTL_IF_START:
            if(unlikely(get_user(word,(uint16_t __user*)compat_ptr(arg))<0))
            {
                return -EFAULT;
            }

            status=g35_if_start(dev,word);
            break;

        case G35_IOCTL_IF_STOP:
            g35_if_stop(dev);

            status=0;
            break;

        case G35_IOCTL_GET_STATE:
            status=g35_get_device_state(dev,compat_ptr(arg));
            break;

        case G35_IOCTL_SET_NOISE_BLANKER:
            if(unlikely(get_user(word,(uint16_t __user*)compat_ptr(arg))<0))
            {
                return -EFAULT;
            }

            status=g35_set_noise_blanker(dev,word);
            break;

        case G35_IOCTL_DDC1_0_CANCEL_BUFFERS:
            g35_cancel_buffers(dev,G35_IO_STREAM_IN_DDC);
            status=0;
            break;

        case G35_IOCTL_DDC1_1_CANCEL_BUFFERS:
            g35_cancel_buffers(dev,G35_IO_STREAM_IN_DDC_WIDE);
            status=0;
            break;

        case G35_IOCTL_CANCEL_BUFFERS_PLAYBACK_IN:
            g35_cancel_buffers(dev,G35_IO_STREAM_IN_DDC_PLAYBACK);
            status=0;
            break;

        case G35_IOCTL_CANCEL_BUFFERS_PLAYBACK_OUT:
            g35_cancel_buffers(dev,G35_IO_STREAM_OUT_DDC_PLAYBACK);
            status=0;
            break;

        case G35_IOCTL_SET_POWER:
            if(unlikely(get_user(byte,(uint8_t __user*)compat_ptr(arg))<0))
            {
                return -EFAULT;
            }

            status=g35_set_power(dev,byte,1);
            break;

        case G35_IOCTL_CONFIGURE_FPGA_COMPAT:
            if(unlikely(copy_from_user(&fpga,(struct g35_fpga_data __user*)compat_ptr(arg),sizeof(fpga))!=0))
            {
                return -EFAULT;
            }

            status=g35_configure_fpga(dev,(void __user*)compat_ptr(fpga.buffer),fpga.size);
            break;

        case G35_IOCTL_GET_CONFIG:
            status=copy_to_user((void __user*)compat_ptr(arg),dev->config,sizeof(dev->config))?-EFAULT:0;
            break;

        case G35_IOCTL_WAIT_FOR_POWER_OFF:
            status=g35_wait_for_power_off(dev);
            break;

        case G35_IOCTL_RF_CONTROL1:
            if(unlikely(copy_from_user(&rf,(union g35_rf_control_compat __user*)compat_ptr(arg),sizeof(rf))!=0))
            {
                return -EFAULT;
            }

            status=g35_rf_control1(dev,rf.value);
            break;
        case G35_IOCTL_RF_CONTROL2:
            if(unlikely(copy_from_user(&rf,(union g35_rf_control_compat __user*)compat_ptr(arg),sizeof(rf))!=0))
            {
                return -EFAULT;
            }

            status=g35_rf_control2(dev,rf.value);
            break;

        case G35_IOCTL_DDC1_0_GET_TIMESTAMP:
            status=g35_ddc1_get_timestamp(dev,G35_DDC_NARROW,compat_ptr(arg));
            break;

        case G35_IOCTL_DDC1_1_GET_TIMESTAMP:
            status=g35_ddc1_get_timestamp(dev,G35_DDC_WIDE,compat_ptr(arg));
            break;

        case G35_IOCTL_COH_ARM:
            status=g35_coh_arm(dev,compat_ptr(arg));
            break;

        case G35_IOCTL_COH_TRIGGER:
            status=g35_coh_trigger(dev);
            break;

        case G35_IOCTL_COH_SET:
            status=g35_coh_set(dev,compat_ptr(arg));
            break;

        case G35_IOCTL_COH_GET:
            status=g35_coh_get(dev,compat_ptr(arg));
            break;

        default:
            status=-EINVAL;
            break;
    }

    return status;
}
#endif //defined(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL)

module_init(g35_init);
module_exit(g35_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("RADIXON");
MODULE_DESCRIPTION(G35_DESCRIPTION);
